from kliff.utils import torch_available

from .calculator import Calculator

__all__ = ["Calculator"]


if torch_available():
    from .calculator_torch import CalculatorTorch

    __all__.extend(["CalculatorTorch"])



import sys
from typing import Union

from loguru import logger

LOG_LEVEL = None


def set_logger(level: str = "INFO", stderr: bool = True):
    """
    Set up loguru loggers.

    The default logger has a log level of `INFO` and logs to stderr. Call this function
    at the beginning of your script to override it.

    Args:
        level: log level, e.g. DEBUG, INFO, WARNING, ERROR, and CRITICAL.
        stderr: whether to log to  stderr.
    """
    global LOG_LEVEL
    LOG_LEVEL = level

    file_handler = {"sink": "kliff.log", "level": level}
    stderr_handler = {"sink": sys.stderr, "level": level}

    if stderr:
        config = {"handlers": [stderr_handler, file_handler]}
    else:
        config = {"handlers": [file_handler]}

    logger.configure(**config)


def get_log_level() -> Union[None, str]:
    """
    Get the current log level.

    Returns:
        Log level in str, one of DEBUG, INFO, WARNING, ERROR, or CRITICAL.
    """

    return LOG_LEVEL



from pathlib import Path

import numpy as np

from kliff.dataset import Dataset
from kliff.dataset.weight import MagnitudeInverseWeight, Weight

np.random.seed(2022)


def test_base_weight():
    # Base weight class
    # Pick random weights
    cw, ew, fw, sw = np.random.uniform(low=0.0, high=1.1, size=4)
    # Instantiate base weight class
    weight = Weight(
        config_weight=cw, energy_weight=ew, forces_weight=fw, stress_weight=sw
    )

    assert weight.config_weight == cw
    assert weight.energy_weight == ew
    assert weight.forces_weight == fw
    assert weight.stress_weight == sw


def test_magnitude_inverse_weight():
    # Inverse magnitude weight
    # Pick random weight parameters
    c1_e, c2_e = np.random.uniform(low=0.0, high=1.0, size=2)
    c1_f, c2_f = np.random.uniform(low=0.0, high=1.0, size=2)
    c1_s, c2_s = np.random.uniform(low=0.0, high=1.0, size=2)
    # Instantiate the weight class
    weight = MagnitudeInverseWeight(
        weight_params={
            "energy_weight_params": [c1_e, c2_e],
            "forces_weight_params": [c1_f, c2_f],
            "stress_weight_params": [c1_s, c2_s],
        }
    )

    # Load the dataset and set the weight
    # We choose the following data set because it has energy, forces, and stress data
    dataset_path = Path(__file__).parents[1].joinpath("configs_extxyz/Si.xyz")
    tset = Dataset(dataset_path, weight=weight)
    configs = tset.get_configs()

    # Check if my implementation works. I do this by comparing it with my manual
    # calculation.
    for conf in configs:
        # Assert energy weight
        ew_implement = conf.weight.energy_weight  # Retrieve weight
        # Compute weight manually
        energy_norm = np.abs(conf.energy)
        ew_manual = _compute_magnitude_inverse_weight(c1_e, c2_e, energy_norm)

        assert isinstance(ew_manual, float)  # The weight is a float
        # Check the values
        assert np.allclose(ew_implement, ew_manual, rtol=1e-12, atol=1e-12)

        # Assert forces weight
        fw_implement = conf.weight.forces_weight  # Retrieve weight
        # Compute weight manually
        forces_norm = np.repeat(np.linalg.norm(conf.forces, axis=1), 3)
        fw_manual = _compute_magnitude_inverse_weight(c1_f, c2_f, forces_norm)

        assert isinstance(fw_manual, np.ndarray)  # The weight is an array
        assert len(fw_manual) == len(conf.forces.flatten())  # Check the length
        # Check the values
        assert np.allclose(fw_implement, fw_manual, rtol=1e-12, atol=1e-12)

        # Assert stress weight
        sw_implement = conf.weight.stress_weight  # Retrieve weight
        # Compute weight manually
        stress_norm = np.sqrt(
            np.linalg.norm(conf.stress) ** 2 + np.linalg.norm(conf.stress[3:]) ** 2
        )
        sw_manual = _compute_magnitude_inverse_weight(c1_s, c2_s, stress_norm)

        assert isinstance(sw_manual, float)  # The weight is a float
        # Check the values
        assert np.allclose(sw_implement, sw_manual, rtol=1e-12, atol=1e-12)


def _compute_magnitude_inverse_weight(c1, c2, norm):
    """Compute the inverse magnitude weight, given the weight parameters and the
    magnitude of property of interest.
    """
    sigma = np.sqrt(c1**2 + (c2 * norm) ** 2)
    return 1 / sigma



"""
.. _tut_mcmc:

MCMC sampling
=============

In this example, we demonstrate how to perform uncertainty quantification (UQ) using
parallel tempered MCMC (PTMCMC). We use a Stillinger-Weber (SW) potential for silicon
that is archived in OpenKIM_.

For simplicity, we only set the energy-scaling parameters, i.e., ``A`` and ``lambda`` as
the tunable parameters. Furthermore, these parameters are physically constrained to be
positive, thus we will work in log parameterization, i.e. ``log(A)`` and ``log(lambda)``.
These parameters will be calibrated to energies and forces of a small dataset,
consisting of 4 compressed and stretched configurations of diamond silicon structure.
"""


##########################################################################################
# To start, let's first install the SW model::
#
#    $ kim-api-collections-management install user SW_StillingerWeber_1985_Si__MO_405512056662_006
#
# .. seealso::
#    This installs the model and its driver into the ``User Collection``. See
#    :ref:`install_model` for more information about installing KIM models.


from multiprocessing import Pool

import numpy as np
from corner import corner

from kliff.calculators import Calculator
from kliff.dataset import Dataset
from kliff.dataset.weight import MagnitudeInverseWeight
from kliff.loss import Loss
from kliff.models import KIMModel
from kliff.models.parameter_transform import LogParameterTransform
from kliff.uq import MCMC, autocorr, mser, rhat
from kliff.utils import download_dataset

##########################################################################################
# Before running MCMC, we need to define a loss function and train the model. More detail
# information about this step can be found in :ref:`tut_kim_sw` and
# :ref:`tut_params_transform`.

# Instantiate a transformation class to do the log parameter transform
param_names = ["A", "lambda"]
params_transform = LogParameterTransform(param_names)

# Create the model
model = KIMModel(
    model_name="SW_StillingerWeber_1985_Si__MO_405512056662_006",
    params_transform=params_transform,
)

# Set the tunable parameters and the initial guess
opt_params = {
    "A": [["default", -8.0, 8.0]],
    "lambda": [["default", -8.0, 8.0]],
}

model.set_opt_params(**opt_params)
model.echo_opt_params()

# Get the dataset and set the weights
dataset_path = download_dataset(dataset_name="Si_training_set_4_configs")
# Instantiate the weight class
weight = MagnitudeInverseWeight(
    weight_params={
        "energy_weight_params": [0.0, 0.1],
        "forces_weight_params": [0.0, 0.1],
    }
)
# Read the dataset and compute the weight
tset = Dataset(dataset_path, weight=weight)
configs = tset.get_configs()

# Create calculator
calc = Calculator(model)
ca = calc.create(configs)

# Instantiate the loss function
residual_data = {"normalize_by_natoms": False}
loss = Loss(calc, residual_data=residual_data)

# Train the model
loss.minimize(method="L-BFGS-B", options={"disp": True})
model.echo_opt_params()


##########################################################################################
# To perform MCMC simulation, we use :class:`~kliff.uq.MCMC`.This class interfaces with
# ptemcee_ Python package to run PTMCMC, which utilizes the affine invariance property
# of MCMC sampling. We simulate MCMC sampling at several different temperatures to
# explore the effect of the scale of bias and overall error bars.

# Define some variables that correspond to the dimensionality of the problem
ntemps = 4  # Number of temperatures to simulate
ndim = calc.get_num_opt_params()  # Number of parameters
nwalkers = 2 * ndim  # Number of parallel walkers to simulate


##########################################################################################
# We start by instantiating :class:`~kliff.uq.MCMC`. This requires :class:`~kliff.loss.Loss`
# instance to construct the likelihood function. Additionally, we can specify the prior
# (or log-prior to be more precise) via the ``logprior_fn`` argument, with the default
# option be a uniform prior that is bounded over a finite range that we specify via the
# ``logprior_args`` argument.
#
# .. note::
#    When user uses the default uniform prior but doesn't specify the bounds, then the
#    sampler will retrieve the bounds from the model
#    (see :meth:`~kliff.models.KIMModel.set_opt_params`). Note that an error will be
#    raised when the uniform prior extends to infinity in any parameter direction.
#
# To specify the sampling temperatures to use, we can use the arguments ``ntemps`` and
# ``Tmax_ratio`` to set how many temperatures to simulate and the ratio of the highest
# temperature to the natural temperature :math:`T_0`, respectively. The default values of
# ``ntemps`` and ``Tmax_ratio`` are 10 and 1.0, respectively. Then, an internal function
# will create a list of logarithmically spaced points from :math:`T = 1.0` to
# :math:`T = T_{\text{max\_ratio}} \times T_0`. Alternatively, we can also give a list of
# the temperatures via ``Tladder`` argument, which will overwrites ``ntemps`` and
# ``Tmax_ratio``.
#
# .. note::
#    It has been shown that including temperatures higher than :math:`T_0` helps the
#    convergence of walkers sampled at :math:`T_0`.
#
# The sampling processes can be parallelized by specifying the pool. Note that the pool
# needs to be declared after instantiating :class:`~kliff.uq.MCMC`, since the posterior
# function is defined during this process.


# Set the boundaries of the uniform prior
bounds = np.tile([-8.0, 8.0], (ndim, 1))

# It is a good practice to specify the random seed to use in the calculation to generate
# a reproducible simulation.
seed = 1717
np.random.seed(seed)

# Create a sampler
sampler = MCMC(
    loss,
    ntemps=ntemps,
    logprior_args=(bounds,),
    random=np.random.RandomState(seed),
)
# Declare a pool to use parallelization
sampler.pool = Pool(nwalkers)


##########################################################################################
# .. note::
#    As a default, the algorithm will set the number of walkers for each sampling
#    temperature to be twice the number of parameters, but we can also specify it via
#    the ``nwalkers`` argument.
#
# To run the MCMC sampling, we use :meth:`~kliff.uq.MCMC.run_mcmc`. This function requires
# us to provide initial states :math:`p_0` for each temperature and walker. We also need
# to specify the number of steps or iterations to take.
#
# .. note::
#    The initial states :math:`p_0` need to be an array with shape ``(K, L, N,)``, where
#    ``K``, ``L``, and ``N`` are the number of temperatures, walkers, and parameters,
#    respectively.


# Initial starting point. This should be provided by the user.
p0 = np.empty((ntemps, nwalkers, ndim))
for ii, bound in enumerate(bounds):
    p0[:, :, ii] = np.random.uniform(*bound, (4, 4))

# Run MCMC
sampler.run_mcmc(p0, 5000)
sampler.pool.close()

# Retrieve the chain
chain = sampler.chain


##########################################################################################
# The resulting chains still need to be processed. First, we need to discard the first few
# iterations in the beginning of each chain as a burn-in time. This is similar to the
# equilibration time in a molecular dynamic simulation before we can start the
# measurement. KLIFF provides a function to estimate the burn-in time, based on the
# Marginal Standard Error Rule (MSER). This can be accessed via
# :func:`~kliff.uq.mcmc_utils.mser`.


# Estimate equilibration time using MSER for each temperature, walker, and dimension.
mser_array = np.empty((ntemps, nwalkers, ndim))
for tidx in range(ntemps):
    for widx in range(nwalkers):
        for pidx in range(ndim):
            mser_array[tidx, widx, pidx] = mser(
                chain[tidx, widx, :, pidx], dmin=0, dstep=10, dmax=-1
            )

burnin = int(np.max(mser_array))
print(f"Estimated burn-in time: {burnin}")


##########################################################################################
# .. note::
#    :func:`~kliff.uq.mcmc_utils.mser` only compute the estimation of the burn-in time for
#    one single temperature, walker, and parameter. Thus, we need to calculate the burn-in
#    time for each temperature, walker, and parameter separately.
#
# After discarding the first few iterations as the burn-in time, we only want to keep
# every :math:`\tau`-th iteration from the remaining chain, where :math:`\tau` is the
# autocorrelation length, to ensure uncorrelated samples.
# This calculation can be done using :func:`~kliff.uq.mcmc_utils.autocorr`.


# Estimate the autocorrelation length for each temperature
chain_no_burnin = chain[:, :, burnin:]

acorr_array = np.empty((ntemps, nwalkers, ndim))
for tidx in range(ntemps):
    acorr_array[tidx] = autocorr(chain_no_burnin[tidx], c=1, quiet=True)

thin = int(np.ceil(np.max(acorr_array)))
print(f"Estimated autocorrelation length: {thin}")


##########################################################################################
# .. note::
#    :func:`~kliff.uq.mcmc_utils.acorr` is a wrapper for emcee.autocorr.integrated_time_,
#    As such, the shape of the input array for this function needs to be ``(L, M, N,)``,
#    where ``L``, ``M``, and ``N`` are the number of walkers, steps, and parameters,
#    respectively. This also implies that we need to perform the calculation for each
#    temperature separately.
#
# Finally, after obtaining the independent samples, we need to assess whether the
# resulting samples have converged to a stationary distribution, and thus a good
# representation of the actual posterior. This is done by computing the potential scale
# reduction factor (PSRF), denoted by :math:`\hat{R}^p`. The value of :math:`\hat{R}^p`
# declines to 1 as the number of iterations goes to infinity. A common threshold is about
# 1.1, but higher threshold has also been used.


# Assess the convergence for each temperature
samples = chain_no_burnin[:, :, ::thin]

threshold = 1.1  # Threshold for rhat
rhat_array = np.empty(ntemps)
for tidx in range(ntemps):
    rhat_array[tidx] = rhat(samples[tidx])

print(f"$\hat{{r}}^p$ values: {rhat_array}")


##########################################################################################
# .. note::
#    :func:`~kliff.uq.mcmc_utils.rhat` only computes the PSRF for one temperature, so that
#    the calculation needs to be carried on for each temperature separately.
#
# Notice that in this case, :math:`\hat{R}^p < 1.1` for all temperatures. When this
# criteria is not satisfied, then the sampling process should be continued. Note that
# some sampling temperatures might converge at slower rates compared to the others.
#
# After obtaining the independent samples from the MCMC sampling, the uncertainty of the
# parameters can be obtained by observing the distribution of the samples. As an example,
# we will use corner_ Python package to present the MCMC result at sampling
# temperature 1.0 as a corner plot.

# Plot samples at T=1.0
corner(samples[0].reshape((-1, ndim)), labels=[r"$\log(A)$", r"$\log(\lambda)$"])

##########################################################################################
# .. note::
#    As an alternative, KLIFF also provides a wrapper to emcee_. This can be accessed by
#    setting ``sampler="emcee"`` when instantiating :class:`~kliff.uq.MCMC`. For further
#    documentation, see :class:`~kliff.uq.EmceeSampler`.
#
# .. _OpenKIM: https://openkim.org
# .. _ptemcee: https://github.com/willvousden/ptemcee
# .. _emcee: https://emcee.readthedocs.io
# .. _emcee.autocorr.integrated_time: https://emcee.readthedocs.io/en/stable/user/autocorr/#emcee.autocorr.integrated_time
# .. _corner: https://corner.readthedocs.io



import argparse
import textwrap
from importlib import import_module

from kliff import __version__

commands = [("dataset", "kliff.cmdline.dataset"), ("model", "kliff.cmdline.model")]


def main():
    parser = argparse.ArgumentParser(description="KLIFF utility collections.")
    parser.add_argument(
        "-v", "--version", action="version", version="KLIFF {}".format(__version__)
    )

    # sub-command
    subparsers = parser.add_subparsers(title="Sub-commands", dest="command")

    # help
    subparser = subparsers.add_parser(
        "help", description="Help", help="Help for sub-command."
    )
    subparser.add_argument(
        "helpcommand",
        nargs="?",
        metavar="sub-command",
        help="Provide help for sub-command.",
    )
    # add all subparser
    parsers = dict()
    functions = dict()
    for cmd_name, mod_name in commands:
        command = import_module(mod_name).Command
        docstring = command.__doc__
        parts = docstring.split("\n", 1)
        if len(parts) == 1:
            short = long = textwrap.dedent(docstring)
        else:
            short, body = parts
            long = short + "\n" + textwrap.dedent(body)

        subparser = subparsers.add_parser(cmd_name, help=short, description=long)
        command.add_arguments(subparser)
        parsers[cmd_name] = subparser
        functions[cmd_name] = command.run

    args = parser.parse_args()
    cmd_name = args.command

    if cmd_name is None:
        parser.print_help()
    elif cmd_name == "help":
        if args.helpcommand is None:
            parser.print_help()
        else:
            parsers[args.helpcommand].print_help()
    else:
        run = functions[cmd_name]
        try:
            if run.__code__.co_argcount == 1:
                run(args)
            else:
                run(args, parsers[cmd_name])
        except KeyboardInterrupt:
            pass
        except Exception as e:
            parser.error("{}: {}\n".format(e.__class__.__name__, e))


if __name__ == "__main__":
    main()



from pathlib import Path

import numpy as np
import pytest

from kliff.dataset import Dataset
from kliff.models.kim import KIMComputeArguments, KIMModel
from kliff.models.parameter_transform import LogParameterTransform

ref_energies = [-277.409737571, -275.597759276, -276.528342759, -275.482988187]


ref_forces = [
    [
        [-1.15948917e-14, -1.15948917e-14, -1.16018306e-14],
        [-1.16989751e-14, -3.93232367e-15, -3.93232367e-15],
        [-3.92538477e-15, -1.17128529e-14, -3.92538477e-15],
    ],
    [
        [1.2676956, 0.1687802, -0.83520474],
        [-1.2536342, 1.19394052, -0.75371034],
        [-0.91847129, -0.26861574, -0.49488973],
    ],
    [
        [0.18042083, -0.11940541, 0.01800594],
        [0.50030564, 0.09165797, -0.09694234],
        [-0.23992404, -0.43625564, 0.14952855],
    ],
    [
        [-1.1114163, 0.21071302, 0.55246303],
        [1.51947195, -1.21522541, 0.7844481],
        [0.54684859, 0.01018317, 0.5062204],
    ],
]


def test_compute():

    # model
    modelname = "SW_StillingerWeber_1985_Si__MO_405512056662_006"
    model = KIMModel(modelname)

    # training set
    path = Path(__file__).parents[1].joinpath("configs_extxyz/Si_4")
    data = Dataset(path)
    configs = data.get_configs()

    # compute arguments
    compute_arguments = []
    for conf in configs:
        ca = model.create_a_kim_compute_argument()
        compute_arguments.append(
            KIMComputeArguments(
                ca,
                conf,
                supported_species=model.get_supported_species(),
                influence_distance=model.get_influence_distance(),
            )
        )
    for i, ca in enumerate(compute_arguments):
        ca.compute(model.kim_model)
        energy = ca.get_energy()
        forces = ca.get_forces()[:3]

        assert energy == pytest.approx(ref_energies[i], 1e-6)
        assert np.allclose(forces, ref_forces[i])


def test_set_one_param():
    modelname = "SW_StillingerWeber_1985_Si__MO_405512056662_006"
    model = KIMModel(modelname)

    # parameters
    params = model.get_model_params()

    # keep the original copy
    sigma = params["sigma"][0]

    model.set_one_opt_param(name="sigma", settings=[[sigma + 0.1]])

    assert params["sigma"][0] == sigma + 0.1

    # internal kim params
    kim_params = model.get_kim_model_params()
    assert kim_params["sigma"][0] == sigma + 0.1


def test_set_opt_params():
    modelname = "SW_StillingerWeber_1985_Si__MO_405512056662_006"
    model = KIMModel(modelname)

    # parameters
    params = model.get_model_params()
    sigma = params["sigma"][0]
    A = params["A"][0]

    model.set_opt_params(sigma=[[sigma + 0.1]], A=[[A + 0.1]])

    assert params["sigma"][0] == sigma + 0.1
    assert params["A"][0] == A + 0.1

    # internal kim params
    kim_params = model.get_kim_model_params()
    assert kim_params["sigma"][0] == sigma + 0.1
    assert kim_params["A"][0] == A + 0.1


def test_get_update_params():
    modelname = "SW_StillingerWeber_1985_Si__MO_405512056662_006"

    model = KIMModel(modelname)

    # parameters
    params = model.get_model_params()
    sigma = params["sigma"][0]
    A = params["A"][0]

    # optimizing parameters
    # B will not be optimized, only providing initial guess
    model.set_opt_params(sigma=[["default"]], B=[["default", "fix"]], A=[["default"]])

    x0 = model.get_opt_params()
    assert x0[0] == sigma
    assert x0[1] == A
    assert len(x0) == 2
    assert model.get_num_opt_params() == 2

    x1 = [i + 0.1 for i in x0]
    model.update_model_params(x1)

    assert params["sigma"][0] == sigma + 0.1
    assert params["A"][0] == A + 0.1

    # internal kim params
    kim_params = model.get_kim_model_params()
    assert kim_params["sigma"][0] == sigma + 0.1
    assert kim_params["A"][0] == A + 0.1


def test_params_transform():

    modelname = "SW_StillingerWeber_1985_Si__MO_405512056662_006"
    model = KIMModel(
        modelname, params_transform=LogParameterTransform(param_names=["sigma", "A"])
    )

    # reference values in KIM model
    sigma = 2.0951
    A = 15.2848479197914
    B = 0.6022245584

    # Check forward transform (all in original space)
    assert model.model_params["sigma"][0] == sigma
    assert model.model_params["A"][0] == A
    assert model.model_params["B"][0] == B

    # Transformed params in log space
    # No log for B since it is not asked to transform
    assert model.model_params_transformed["sigma"][0] == np.log(sigma)
    assert model.model_params_transformed["A"][0] == np.log(A)
    assert model.model_params_transformed["B"][0] == B

    # optimizing parameters, provided in log space
    # B will not be optimized, only providing initial guess
    v1 = 2.0
    v2 = 3.0
    model.set_opt_params(sigma=[[v1]], B=[[B, "fix"]], A=[[v2]])

    assert model.model_params_transformed["sigma"][0] == v1
    assert model.model_params_transformed["A"][0] == v2
    assert model.model_params_transformed["B"][0] == B

    assert np.allclose(model.get_opt_params(), [v1, v2])

    model.update_model_params([v1, v2])

    # Check inverse transform
    kim_params = model.get_kim_model_params()
    assert kim_params["sigma"].value[0] == np.exp(v1)
    assert kim_params["A"].value[0] == np.exp(v2)
    assert kim_params["B"].value[0] == B



from pathlib import Path
from typing import Dict, List, Optional, Union

import torch
from loguru import logger
from torch.utils.data import DataLoader

from kliff.dataset.dataset import Configuration
from kliff.dataset.dataset_torch import FingerprintsDataset, fingerprints_collate_fn
from kliff.models.model_torch import ModelTorch
from kliff.models.neural_network import NeuralNetwork
from kliff.utils import pickle_load, to_path


class CalculatorTorch:
    """
    A calculator for torch based models.

    Args:
        model: torch models, e.g. :class:`~kliff.neuralnetwork.NeuralNetwork`.
        gpu: whether to use gpu for training. If `int` (e.g. 0), will trained on this
            gpu device. If `True` will always train on gpu `0`.
    """

    implemented_property = ["energy", "forces", "stress"]

    def __init__(self, model: ModelTorch, gpu: Union[bool, int] = None):
        device = _get_device(gpu)
        self._model = model.to(device)

        self.dtype = self.model.descriptor.dtype
        self.fingerprints_path = None

        self.use_energy = None
        self.use_forces = None
        self.use_stress = None

        self.results = dict([(i, None) for i in self.implemented_property])

    def create(
        self,
        configs: List[Configuration],
        use_energy: bool = True,
        use_forces: bool = True,
        use_stress: bool = False,
        fingerprints_filename: Union[Path, str] = "fingerprints.pkl",
        fingerprints_mean_stdev_filename: Optional[Union[Path, str]] = None,
        reuse: bool = False,
        use_welford_method: bool = False,
        nprocs: int = 1,
    ):
        """
        Process configs to generate fingerprints.

        Args:
            configs: atomic configurations
            use_energy: Whether to require the calculator to compute energy.
            use_forces: Whether to require the calculator to compute forces.
            use_stress: Whether to require the calculator to compute stress.
            fingerprints_filename: Path to save the generated fingerprints.
                If `reuse=True`, Will not generate the fingerprints, but directly use the
                one provided via this file.
            fingerprints_mean_stdev_filename: Path to save the mean and standard deviation
                of the fingerprints. If `reuse=True`, Will not generate new fingerprints
                mean and stdev, but directly use the one provided via this file.
                If `normalize` is not required by a descriptor, this is ignored.
            reuse: Whether to reuse provided fingerprints.
            use_welford_method: Whether to compute mean and standard deviation using the
                Welford method, which is memory efficient. See
                https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance
            nprocs: Number of processes used to generate the fingerprints. If `1`, run
                in serial mode, otherwise `nprocs` processes will be forked via
                multiprocessing to do the work.
        """

        self.configs = configs
        self.use_energy = use_energy
        self.use_forces = use_forces
        self.use_stress = use_stress

        if isinstance(configs, Configuration):
            configs = [configs]

        # reuse existing file
        if reuse:
            self.fingerprints_path = to_path(fingerprints_filename)
            if not self.fingerprints_path.exists():
                raise CalculatorTorchError(
                    f"You specified `reuse=True` to reuse the fingerprints stored in "
                    f"`{self.fingerprints_path}` This file does not exists."
                )
            logger.info(f"Reuse fingerprints `{self.fingerprints_path}`")

            if self.model.descriptor.normalize:
                path = (
                    None
                    if fingerprints_mean_stdev_filename is None
                    else to_path(fingerprints_mean_stdev_filename)
                )

                if path is None or not path.exists():
                    raise CalculatorTorchError(
                        f"You specified `reuse=True` to reuse the fingerprints. The "
                        f"mean and stdev file of the fingerprints `{path}` does not "
                        "exists."
                    )

                self.model.descriptor.load_state_dict(pickle_load(path))

                logger.info(f"Reuse fingerprints mean and stdev `{path}`")

        # generate fingerprints and pickle it
        else:
            self.fingerprints_path = self.model.descriptor.generate_fingerprints(
                configs,
                use_forces,
                use_stress,
                fingerprints_filename,
                fingerprints_mean_stdev_filename,
                use_welford_method,
                nprocs,
            )

    def get_compute_arguments(self, batch_size: int = 1):
        """
        Return the dataloader with batch size set to `batch_size`.
        """
        fname = self.fingerprints_path
        fp = FingerprintsDataset(fname)
        loader = DataLoader(
            dataset=fp, batch_size=batch_size, collate_fn=fingerprints_collate_fn
        )

        return loader

    def fit(self):
        path = self.fingerprints_path
        self.model.fit(path)

    def compute(self, batch):

        #
        # shape N--number of atoms in a config; D--feature dim
        # zeta: (N, D)
        # dzetadr_force: (N, D, 3N)
        # dzetadr_stress: (N, D, 6)
        #
        # batching dzetadr_force seems difficult, because two axes have different size
        # this seems doable, combine N and 3N as one dim, and use einstein sum

        device = self.model.device

        grad = self.use_forces or self.use_stress

        # TODO, the batching should be moved to dataloader
        # get information from batch
        zeta_config = [sample["zeta"] for sample in batch]
        zeta_stacked = torch.cat(zeta_config, dim=0).to(device)

        # evaluate model
        if grad:
            zeta_stacked.requires_grad_(True)

        energy_atom = self.model(zeta_stacked)

        # forces and stress
        if not self.use_forces:
            forces_config = None
        else:
            forces_config = []
        if not self.use_stress:
            stress_config = None
        else:
            stress_config = []

        natoms_config = [len(zeta) for zeta in zeta_config]
        energy_config = [e.sum() for e in torch.split(energy_atom, natoms_config)]

        if grad:
            dedzeta = torch.autograd.grad(
                energy_atom.sum(), zeta_stacked, create_graph=True
            )[0]
            zeta_stacked.requires_grad_(False)  # no need of grad any more

            dedzeta_config = torch.split(dedzeta, natoms_config)

            for i, sample in enumerate(batch):
                dedz = dedzeta_config[i]

                if self.use_forces:
                    dzetadr_forces = sample["dzetadr_forces"].to(device)
                    f = self._compute_forces(dedz, dzetadr_forces)
                    forces_config.append(f)

                if self.use_stress:
                    dzetadr_stress = sample["dzetadr_stress"].to(device)
                    volume = sample["dzetadr_volume"]
                    s = self._compute_stress(dedz, dzetadr_stress, volume)
                    stress_config.append(s)

        self.results["energy"] = energy_config
        self.results["forces"] = forces_config
        self.results["stress"] = stress_config
        return {
            "energy": energy_config,
            "forces": forces_config,
            "stress": stress_config,
        }

    @property
    def model(self):
        """Get the underlying torch model"""
        return self._model

    def save_model(self, epoch: int, force_save: bool = False):
        """
        Save the model to disk.

        When to save a model is dependent on `epoch` and a model's metadata for save.

        Args:
            epoch: current optimization epoch.
            force_save: save the model, ignoring `epoch` and save metadata.
        """
        # save metadata
        save_prefix = self.model.save_prefix
        save_start = self.model.save_start
        save_frequency = self.model.save_frequency

        path = to_path(save_prefix).joinpath(f"model_epoch{epoch}.pkl")
        if force_save:
            self.model.save(path)
        else:
            if epoch >= save_start and (epoch - save_start) % save_frequency == 0:
                self.model.save(path)

    def get_energy(self, batch):
        return self.results["energy"]

    def get_forces(self, batch):
        return self.results["forces"]

    def get_stress(self, batch):
        return self.results["stress"]

    @staticmethod
    def _compute_forces(denergy_dzeta, dzetadr):
        forces = -torch.tensordot(denergy_dzeta, dzetadr, dims=([0, 1], [0, 1]))
        return forces

    @staticmethod
    def _compute_stress(denergy_dzeta, dzetadr, volume):
        forces = torch.tensordot(denergy_dzeta, dzetadr, dims=([0, 1], [0, 1])) / volume
        return forces


class CalculatorTorchSeparateSpecies(CalculatorTorch):
    """
    A calculator supporting models of difference species.

    Args:
        models: {species: model} with species specifying the chemical symbol for the
            model.
        gpu: whether to use gpu for training. If `int` (e.g. 0), will trained on this
            gpu device. If `True` will always train on gpu `0`.
    """

    def __init__(self, models: Dict[str, NeuralNetwork], gpu: Union[bool, int] = None):
        device = _get_device(gpu)

        self.models = models

        self.dtype = None
        for s, m in self.models.items():
            m.to(device)

            if self.dtype is None:
                self.dtype = m.descriptor.dtype
            else:
                if self.dtype != m.descriptor.dtype:
                    raise CalculatorTorchError("inconsistent `dtype` from descriptors.")

        self._model = _ModelWrapper(models)

        self.fingerprints_path = None

        self.use_energy = None
        self.use_forces = None
        self.use_stress = None

        self.results = dict([(i, None) for i in self.implemented_property])

    def compute(self, batch):

        device = self.model.device

        grad = self.use_forces or self.use_stress

        # collate batch by species

        supported_species = self.models.keys()
        zeta_by_species = {s: [] for s in supported_species}
        config_id_by_species = {s: [] for s in supported_species}
        zeta_config = []

        for i, sample in enumerate(batch):
            zeta = sample["zeta"].to(device)
            species = sample["configuration"].species
            zeta.requires_grad_(True)
            zeta_config.append(zeta)

            for s, z in zip(species, zeta):
                if s not in supported_species:
                    raise CalculatorTorchError(f"No model for species: {s}")
                else:
                    zeta_by_species[s].append(z)
                    config_id_by_species[s].append(i)

        # evaluate model to compute energy
        energy_config = [None for _ in range(len(batch))]
        for s, zeta in zeta_by_species.items():

            # have no species "s" in this batch of data
            if not zeta:  # zeta == []
                continue

            z_tensor = torch.stack(zeta)  # convert a list of tensor to tensor
            energy = self.models[s](z_tensor)

            for e_atom, i in zip(energy, config_id_by_species[s]):
                if energy_config[i] is None:
                    energy_config[i] = e_atom
                else:
                    # note cannot use +=, energy e_atom is a view
                    energy_config[i] = energy_config[i] + e_atom

        # forces and stress
        if not self.use_forces:
            forces_config = None
        else:
            forces_config = []
        if not self.use_stress:
            stress_config = None
        else:
            stress_config = []
        if grad:
            for i, sample in enumerate(batch):

                # derivative of energy w.r.t. zeta
                energy = energy_config[i]
                zeta = zeta_config[i]
                dedz = torch.autograd.grad(energy, zeta, create_graph=True)[0]
                zeta.requires_grad_(False)  # no need of grad any more

                if self.use_forces:
                    dzetadr_forces = sample["dzetadr_forces"].to(device)
                    f = self._compute_forces(dedz, dzetadr_forces)
                    forces_config.append(f)

                if self.use_stress:
                    dzetadr_stress = sample["dzetadr_stress"]
                    volume = sample["dzetadr_volume"].to(device)
                    s = self._compute_stress(dedz, dzetadr_stress, volume)
                    stress_config.append(s)

        self.results["energy"] = energy_config
        self.results["forces"] = forces_config
        self.results["stress"] = stress_config
        return {
            "energy": energy_config,
            "forces": forces_config,
            "stress": stress_config,
        }

    @property
    def model(self):
        return self._model

    def save_model(self, epoch: int, force_save: bool = False):
        """
        Save the models to disk.

        When to save a model is dependent on `epoch` and a model's metadata for save.

        Args:
            epoch: current optimization epoch.
            force_save: save the model, ignoring `epoch` and save metadata.
        """
        # save metadata
        for name, model in self.models.items():
            save_prefix = model.save_prefix
            save_start = model.save_start
            save_frequency = model.save_frequency

            path = to_path(save_prefix).joinpath(f"model_{name}_epoch{epoch}.pkl")
            if force_save:
                model.save(path)
            else:
                if epoch >= save_start and (epoch - save_start) % save_frequency == 0:
                    model.save(path)


class _ModelWrapper(torch.nn.Module):
    """
    A wrapper over multiple torch models.

    Only add necessary properties:
      - `LossNeuralNetworkModel` uses `calculator.model.parameters()` and
      - `calculator.model.device`, and the model wrapper only need to provide them.
      - descriptor: needed by model create
    """

    def __init__(self, models: Dict[str, torch.nn.Module]):
        super().__init__()
        self._models = torch.nn.ModuleDict(models)

        first_model = list(models.values())[0]

        # Assuming all models using the same descriptor as in the example_NN_SiC.py
        # example, then it's OK to set it to the descriptor of the first model.
        self._descriptor = first_model.descriptor

    @property
    def device(self):
        return next(self.parameters()).device

    @property
    def descriptor(self):
        return self._descriptor


# class CalculatorTorchDDP(CalculatorTorch):
#     def __init__(self, model, rank, world_size):
#         super(self).__init__(model)
#         self.set_up(rank, world_size)
#
#     def set_up(self, rank, world_size):
#         os.environ["MASTER_ADDR"] = "localhost"
#         os.environ["MASTER_PORT"] = "12355"
#         dist.init_process_group("gloo", rank=rank, world_size=world_size)
#
#     def clean_up(self):
#         dist.destroy_process_group()
#
#     def compute(self, batch):
#         grad = self.use_forces
#
#         # collate batch input to NN
#         zeta_config = self._collate(batch, "zeta")
#         if grad:
#             for zeta in zeta_config:
#                 zeta.requires_grad_(True)
#         zeta_stacked = torch.cat(zeta_config, dim=0)
#
#         # evaluate model
#         model = DistributedDataParallel(self.model)
#         energy_atom = model(zeta_stacked)
#
#         # energy
#         natoms_config = [len(zeta) for zeta in zeta_config]
#         energy_config = [e.sum() for e in torch.split(energy_atom, natoms_config)]
#
#         # forces
#         if grad:
#             dzetadr_config = self._collate(batch, "dzetadr")
#             forces_config = self.compute_forces_config(
#                 energy_config, zeta_config, dzetadr_config
#             )
#             for zeta in zeta_config:
#                 zeta.requires_grad_(False)
#         else:
#             forces_config = None
#
#         return {"energy": energy_config, "forces": forces_config}
#
#     def __del__(self):
#         self.clean_up()


class CalculatorTorchError(Exception):
    def __init__(self, msg):
        super(CalculatorTorchError, self).__init__(msg)
        self.msg = msg


def _get_device(gpu):

    device = None
    if isinstance(gpu, bool):
        if gpu:
            device = torch.device(0)
            logger.info(f"Training on gpu")
    elif isinstance(gpu, int):
        device = torch.device(gpu)
        logger.info(f"Training on gpu {gpu}")
    if device is None:
        logger.info("Training on cpu")

    return device



from multiprocessing import Pool
from pathlib import Path

import numpy as np
import pytest

from kliff.calculators import Calculator
from kliff.dataset import Dataset
from kliff.loss import Loss
from kliff.models import KIMModel
from kliff.uq import MCMC, get_T0

try:
    from kliff.uq.mcmc import PtemceeSampler

    ptemcee_avail = True
except ImportError:
    ptemcee_avail = False

try:
    from kliff.uq.mcmc import EmceeSampler

    emcee_avail = True
except ImportError:
    emcee_avail = False


seed = 1717
np.random.seed(seed)

# model
modelname = "SW_StillingerWeber_1985_Si__MO_405512056662_006"
model = KIMModel(modelname)
model.set_opt_params(A=[["default"]])

# training set
path = Path(__file__).absolute().parents[1].joinpath("configs_extxyz/Si_4")
data = Dataset(path)
configs = data.get_configs()

# calculator
calc = Calculator(model)
ca = calc.create(configs)

# loss
loss = Loss(calc)
loss.minimize(method="lm")

# dimensionality
ndim = calc.get_num_opt_params()
nwalkers = 2 * np.random.randint(1, 3)
ntemps = np.random.randint(2, 4)
nsteps = np.random.randint(5, 10)

# samplers
prior_bounds = np.tile([0, 10], (ndim, 1))

if ptemcee_avail:
    ptsampler = MCMC(
        loss,
        ntemps=ntemps,
        nwalkers=nwalkers,
        logprior_args=(prior_bounds,),
        random=np.random.RandomState(seed),
    )
if emcee_avail:
    sampler = MCMC(
        loss, nwalkers=nwalkers, logprior_args=(prior_bounds,), sampler="emcee"
    )


def test_T0():
    """Test if the function to compute T0 works properly. This is done by comparing T0
    computed using the internal function and computed manually.
    """
    # Using internal function
    T0_internal = get_T0(loss)
    # Compute manually
    xopt = calc.get_opt_params()
    T0_manual = 2 * loss._get_loss(xopt) / len(xopt)
    assert T0_internal == T0_manual, "Internal function to compute T0 doesn't work"


def test_MCMC_wrapper():
    """Test if the MCMC wrapper class returns the correct sampler instance."""
    if ptemcee_avail:
        assert (
            type(ptsampler) == PtemceeSampler
        ), "MCMC should return ``PtemceeSampler`` instance"
    if emcee_avail:
        assert (
            type(sampler) == EmceeSampler
        ), "MCMC should return ``EmceeSampler`` instance"


def test_dimensionality():
    """Test the number of temperatures, walkers, steps, and parameters. This is done by
    comparing the shape of the resulting MCMC chains and the variables used to set these
    dimensions.
    """

    # Test for ptemcee wrapper
    if ptemcee_avail:
        p0 = np.random.uniform(0, 10, (ntemps, nwalkers, ndim))
        ptsampler.run_mcmc(p0=p0, iterations=nsteps)
        assert ptsampler.chain.shape == (
            ntemps,
            nwalkers,
            nsteps,
            ndim,
        ), "Dimensionality from the ptemcee wrapper is not right"
    else:
        print("Skip testing ptemcee; ptemcee is not found")

    # Test for emcee wrapper
    if emcee_avail:
        p0 = np.random.uniform(0, 10, (nwalkers, ndim))
        sampler.run_mcmc(initial_state=p0, nsteps=nsteps)
        assert sampler.chain.shape == (
            nwalkers,
            nsteps,
            ndim,
        ), "Dimensionality from the emcee wrapper is not right"
    else:
        print("Skip testing emcee; emcee is not found")


def test_pool_exception():
    """Test if an exception is raised when declaring the pool prior to instantiating
    ``kliff.uq.MCMC``.
    """
    if ptemcee_avail:
        with pytest.raises(ValueError):
            _ = MCMC(
                loss,
                ntemps=ntemps,
                nwalkers=nwalkers,
                logprior_args=(prior_bounds,),
                pool=Pool(1),
            )
    else:
        print("Skip the test; ptemcee is not found")


def test_sampler_exception():
    """Test if an exception is raised when specifying sampler string other than the
    built-in ones.
    """
    with pytest.raises(ValueError):
        _ = MCMC(
            loss,
            ntemps=ntemps,
            nwalkers=nwalkers,
            logprior_args=(prior_bounds,),
            sampler="dummy",
        )


if __name__ == "__main__":
    test_T0()
    test_MCMC_wrapper()
    test_dimensionality()
    test_pool_exception()



import os
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Sequence

import numpy as np
from loguru import logger

from kliff.dataset.dataset import Configuration
from kliff.error import report_import_error
from kliff.log import get_log_level
from kliff.models.model import ComputeArguments, Model
from kliff.models.parameter import Parameter
from kliff.models.parameter_transform import ParameterTransform
from kliff.neighbor import assemble_forces, assemble_stress

try:
    import kimpy
    from kimpy import neighlist as nl

    kimpy_avail = True
except ImportError:
    kimpy_avail = False


class KIMComputeArguments(ComputeArguments):
    """
    KIMModel potentials arguments.

    Args:
        kim_ca: KIM compute argument, can be created by
            :meth:`~kliff.models.KIMModels.create_a_compute_argument()`.
        config: atomic configurations
        supported_species: species supported by the potential model, with chemical
            symbol as key and integer code as value.
        influence_distance: influence distance (aka cutoff distance) to calculate neighbors
        compute_energy: whether to compute energy
        compute_forces: whether to compute forces
        compute_stress: whether to compute stress
    """

    implemented_property = []

    def __init__(
        self,
        kim_ca,
        config: Configuration,
        supported_species: Dict[str, int],
        influence_distance: float,
        compute_energy: bool = True,
        compute_forces: bool = True,
        compute_stress: bool = False,
    ):
        if not kimpy_avail:
            report_import_error("kimpy", self.__class__.__name__)

        # kim compute argument
        self.kim_ca = kim_ca

        # get supported property
        self._get_implemented_property()

        super(KIMComputeArguments, self).__init__(
            config,
            supported_species,
            influence_distance,
            compute_energy,
            compute_forces,
            compute_stress,
        )

        # neighbor list
        self.neigh = None
        self.num_contributing_particles = None
        self.num_padding_particles = None
        self.padding_image_of = None

        # model input
        self.num_particles = None
        self.species_code = None
        self.particle_contributing = None
        self.coords = None

        # model output
        self.energy = None
        self.forces = None

        self._init_neigh()
        self._update_neigh(influence_distance)
        self._register_data(compute_energy, compute_forces)

    def _get_implemented_property(self):
        """
        Get implemented property of model.
        """

        # check compute arguments
        try:
            kim_can = kimpy.compute_argument_name
            N = kim_can.get_number_of_compute_argument_names()
        except RuntimeError:
            raise kimpy.KimPyError(
                "Calling `kim_can.get_number_of_compute_argument_names()` failed."
            )

        for i in range(N):
            try:
                name = kim_can.get_compute_argument_name(i)
            except RuntimeError:
                raise kimpy.KimPyError(
                    "Calling `kim_can.get_compute_argument_name()` failed."
                )

            # dtype = kim_can.get_compute_argument_data_type(name)

            try:
                support_status = self.kim_ca.get_argument_support_status(name)
            except RuntimeError:
                raise kimpy.KimPyError(
                    "Calling `kim_ca.get_argument_support_status()` failed."
                )

            # calculator can only handle energy and forces
            if support_status == kimpy.support_status.required:
                if name != kim_can.partialEnergy and name != kim_can.partialForces:
                    KIMModelError(f"Unsupported required ComputeArgument `{name}`")

            # supported property
            if name == kim_can.partialEnergy:
                self.implemented_property.append("energy")
            elif name == kim_can.partialForces:
                self.implemented_property.append("forces")
                self.implemented_property.append("stress")

        # check compute callbacks
        kim_ccn = kimpy.compute_callback_name

        try:
            num_callbacks = kim_ccn.get_number_of_compute_callback_names()
        except RuntimeError:
            raise kimpy.KimPyError(
                "Calling `kim_ccn.get_number_of_compute_callback_names()` failed."
            )

        for i in range(num_callbacks):
            try:
                name = kim_ccn.get_compute_callback_name(i)
            except RuntimeError:
                raise kimpy.KimPyError(
                    "Calling `kim_ccn.get_compute_callback_name()` failed."
                )

            try:
                support_status = self.kim_ca.get_callback_support_status(name)
            except RuntimeError:
                raise kimpy.KimPyError(
                    "Calling `kim_ca.get_callback_support_status()` failed."
                )

            # calculator only provides get_neigh
            if support_status == kimpy.support_status.required:
                if name != kim_ccn.GetNeighborList:
                    KIMModelError(f"Unsupported required ComputeCallback `{name}`")

    def _init_neigh(self):

        # create neighbor list
        try:
            neigh = nl.create()
        except RuntimeError:
            raise kimpy.KimPyError("Calling `nl.create()` failed.")

        # register get neigh callback
        try:
            self.kim_ca.set_callback_pointer(
                kimpy.compute_callback_name.GetNeighborList, nl.get_neigh_kim(), neigh
            )
        except RuntimeError:
            raise kimpy.KimPyError("Calling `kim_ca.set_callback_pointer()` failed.")

        self.neigh = neigh

    def _update_neigh(self, influence_distance: float):
        """
        Update neighbor list and model input.
        """

        # inquire information from conf
        cell = np.asarray(self.conf.cell, dtype=np.double)
        PBC = np.asarray(self.conf.PBC, dtype=np.intc)
        contributing_coords = np.asarray(self.conf.coords, dtype=np.double)
        contributing_species = self.conf.species
        num_contributing = self.conf.get_num_atoms()
        self.num_contributing_particles = num_contributing

        # species support and code
        unique_species = list(set(contributing_species))
        species_map = dict()
        for s in unique_species:
            if s in self.supported_species:
                species_map[s] = self.supported_species[s]
            else:
                KIMModelError(f"species `{s}` not supported by model")

        contributing_species_code = np.array(
            [species_map[s] for s in contributing_species], dtype=np.intc
        )

        if any(PBC):  # need padding atoms

            try:
                (
                    padding_coords,
                    padding_species_code,
                    self.padding_image_of,
                ) = nl.create_paddings(
                    influence_distance,
                    cell,
                    PBC,
                    contributing_coords,
                    contributing_species_code,
                )
            except RuntimeError:
                raise kimpy.KimPyError("Calling `nl.create_paddings()` failed.")

            num_padding = padding_species_code.size
            self.num_particles = np.array(
                [num_contributing + num_padding], dtype=np.intc
            )
            tmp = np.concatenate((contributing_coords, padding_coords))
            self.coords = np.asarray(tmp, dtype=np.double)
            tmp = np.concatenate((contributing_species_code, padding_species_code))
            self.species_code = np.asarray(tmp, dtype=np.intc)
            self.particle_contributing = np.ones(self.num_particles[0], dtype=np.intc)
            self.particle_contributing[num_contributing:] = 0

            # for now, create neigh for all atoms, including paddings
            need_neigh = np.ones(self.num_particles[0], dtype=np.intc)

        else:  # do not need padding atoms
            self.padding_image_of = np.array([])
            self.num_particles = np.array([num_contributing], dtype=np.intc)
            self.coords = np.array(contributing_coords, dtype=np.double)
            self.species_code = np.array(contributing_species_code, dtype=np.intc)
            self.particle_contributing = np.ones(num_contributing, dtype=np.intc)
            need_neigh = self.particle_contributing

        try:
            self.neigh.build(
                self.coords,
                influence_distance,
                np.asarray([influence_distance], dtype=np.double),
                need_neigh,
            )
        except RuntimeError:
            raise kimpy.KimPyError("Calling `neigh.build()` failed.")

    def _register_data(self, compute_energy=True, compute_forces=True):
        """
        Register model input and output data in KIM API.
        """

        # check whether model support energy and forces
        kim_can = kimpy.compute_argument_name
        if compute_energy:
            name = kim_can.partialEnergy

            try:
                support_status = self.kim_ca.get_argument_support_status(name)
            except RuntimeError:
                raise kimpy.KimPyError(
                    "Calling `kim_ca.get_argument_support_status()` failed."
                )

            if not (
                support_status == kimpy.support_status.required
                or support_status == kimpy.support_status.optional
            ):
                KIMModelError("Energy not supported by model")

        if compute_forces:
            name = kim_can.partialForces

            try:
                support_status = self.kim_ca.get_argument_support_status(name)
            except RuntimeError:
                raise kimpy.KimPyError(
                    "Calling `kim_ca.get_argument_support_status()` failed."
                )

            if not (
                support_status == kimpy.support_status.required
                or support_status == kimpy.support_status.optional
            ):
                KIMModelError("Forces not supported by model")

        # register argument
        try:
            self.kim_ca.set_argument_pointer(
                kim_can.numberOfParticles, self.num_particles
            )

            self.kim_ca.set_argument_pointer(
                kim_can.particleSpeciesCodes, self.species_code
            )

            self.kim_ca.set_argument_pointer(
                kim_can.particleContributing, self.particle_contributing
            )

            self.kim_ca.set_argument_pointer(kim_can.coordinates, self.coords)
        except RuntimeError:
            raise kimpy.KimPyError("Calling `kim_ca.set_argument_pointer()` failed.")

        if compute_energy:
            self.energy = np.array([0.0], dtype=np.double)
            try:
                self.kim_ca.set_argument_pointer(kim_can.partialEnergy, self.energy)
            except RuntimeError:
                raise kimpy.KimPyError(
                    "Calling `kim_ca.set_argument_pointer()` failed."
                )

        else:
            self.energy = None
            try:
                self.kim_ca.set_argument_null_pointer(kim_can.partialEnergy)
            except RuntimeError:
                raise kimpy.KimPyError(
                    "Calling `kim_ca.set_argument_null_pointer()` failed."
                )

        if compute_forces:
            self.forces = np.zeros([self.num_particles[0], 3], dtype=np.double)
            try:
                self.kim_ca.set_argument_pointer(kim_can.partialForces, self.forces)
            except RuntimeError:
                raise kimpy.KimPyError(
                    "Calling `kim_ca.set_argument_pointer()` failed."
                )

        else:
            self.forces = None
            try:
                self.kim_ca.set_argument_null_pointer(kim_can.partialForces)
            except RuntimeError:
                raise kimpy.KimPyError(
                    "Calling `kim_ca.set_argument_null_pointer()` failed."
                )

    def compute(self, kim_model):

        try:
            kim_model.compute(self.kim_ca)
        except RuntimeError:
            raise kimpy.KimPyError("Calling `kim_model.compute()` failed.")

        if self.compute_energy:
            self.results["energy"] = self.energy[0]
        if self.compute_forces:
            forces = assemble_forces(
                self.forces, self.num_contributing_particles, self.padding_image_of
            )
            self.results["forces"] = forces
        if self.compute_stress:
            volume = self.conf.get_volume()
            stress = assemble_stress(self.coords, self.forces, volume)
            self.results["stress"] = stress


class KIMModel(Model):
    """
    A general interface to any KIM model.

    Args:
        model_name: name of a KIM model. Available models can be found at:
            https://openkim.org.
            For example `SW_StillingerWeber_1985_Si__MO_405512056662_006`.
        params_transform: optional transformation of parameters. Let's call the
            parameters initialized in `init_model_params()` as original parameter
            space. Sometimes, it's easier to work in another parameter (e.g.
            optimizers can perform better in the log space). Then we can use this
            parameter transformation class to transform between the original
            parameter space and the new easy-to-work space. Typically, a model only
            knows how to work in its original space to compute, e.g. energy and
            forces, so we need to inverse transform parameters back to original space
            (after an optimizer update its value in the log space).
            A `params_transform` instance should implement both a `transform` and an
            `inverse_transform` method to accomplish the above tasks.
            Note, all the parameters of this (the `Model`) class
            (e.g. `self.model_params`, and `self.opt_params`) are in the transformed
            easy-to-work space.
    """

    def __init__(
        self,
        model_name: str,
        params_transform: Optional[ParameterTransform] = None,
    ):
        if not kimpy_avail:
            report_import_error("kimpy", self.__class__.__name__)

        self.kim_model = self._create_kim_model(model_name)

        super(KIMModel, self).__init__(model_name, params_transform)

    def init_model_params(self) -> Dict[str, Parameter]:
        return self.get_kim_model_params()

    def init_influence_distance(self) -> float:
        return self.kim_model.get_influence_distance()

    def init_supported_species(self) -> Dict[str, int]:
        species = {}

        try:
            num_kim_species = kimpy.species_name.get_number_of_species_names()
        except RuntimeError:
            raise kimpy.KimPyError(
                "Calling `kimpy.species_name.get_number_of_species_names()` failed."
            )

        for i in range(num_kim_species):
            try:
                species_name = kimpy.species_name.get_species_name(i)
            except RuntimeError:
                raise kimpy.KimPyError(
                    "Calling `kimpy.species_name.get_species_name()` failed."
                )

            try:
                supported, code = self.kim_model.get_species_support_and_code(
                    species_name
                )
            except RuntimeError:
                raise kimpy.KimPyError(
                    "Calling `kim_model.get_species_support_and_code()` failed."
                )

            if supported:
                species[str(species_name)] = code

        return species

    def get_compute_argument_class(self):
        return KIMComputeArguments

    @staticmethod
    def _create_kim_model(model_name: str):
        """
        Create a new kim model.
        """
        try:
            units_accepted, model = kimpy.model.create(
                kimpy.numbering.zeroBased,
                kimpy.length_unit.A,
                kimpy.energy_unit.eV,
                kimpy.charge_unit.e,
                kimpy.temperature_unit.K,
                kimpy.time_unit.ps,
                model_name,
            )
        except RuntimeError:
            raise kimpy.KimPyError("Calling `kimpy.model.create()` failed.")

        if not units_accepted:
            KIMModelError("requested units not accepted in kimpy.model.create")
        return model

    def get_kim_model_params(self) -> Dict[str, Parameter]:
        """
        Inquire the KIM model to get all the parameters.

        Returns:
            {name, parameter}, all parameters in a kim model.
        """
        try:
            num_params = self.kim_model.get_number_of_parameters()
        except RuntimeError:
            raise kimpy.KimPyError(
                "Calling `kim_model.get_number_of_parameters()` failed."
            )

        params = dict()
        for i in range(num_params):
            try:
                (
                    dtype,
                    extent,
                    name,
                    description,
                ) = self.kim_model.get_parameter_metadata(i)
            except RuntimeError:
                raise kimpy.KimPyError(
                    "Calling `kim_model.get_parameter_metadata()` failed."
                )

            values = []
            for j in range(extent):
                if str(dtype) == "Double":
                    try:
                        val = self.kim_model.get_parameter_double(i, j)
                    except RuntimeError:
                        raise kimpy.KimPyError(
                            "Calling `kim_model.get_parameter_double()` failed."
                        )
                    values.append(val)

                elif str(dtype) == "Int":
                    try:
                        val = self.kim_model.get_parameter_int(i, j)
                    except RuntimeError:
                        raise kimpy.KimPyError(
                            "Calling `kim_model.get_parameter_int()` failed."
                        )
                    values.append(val)

                else:  # should never reach here
                    KIMModelError(f"get unexpected parameter data type `{dtype}`")

            params[name] = Parameter(value=values, name=name, index=i)

        return params

    def create_a_kim_compute_argument(self):
        """
        Create a compute argument for the KIM model.
        """
        try:
            kim_ca = self.kim_model.compute_arguments_create()
        except RuntimeError:
            raise kimpy.KimPyError(
                "Calling `kim_model.compute_arguments_create()` failed."
            )

        return kim_ca

    def set_opt_params(self, **kwargs):
        """
        Set the parameters that will be optimized.

        One or more parameters can be set. Each argument is for one parameter, where the
        argument name is the parameter name, the value of the argument is the
        settings(including initial value, fix flag, lower bound, and upper bound).

        The value of the argument should be a list of list, where each inner list is for
        one component of the parameter, which can contain 1, 2, or 3 elements.
         See `~kliff.model.model.Model.read_opt_params()` for the options of the elements.

        Example:
           instance.set(A=[['DEFAULT'], [2.0, 1.0, 3.0]], B=[[1.0, 'FIX'], [2.0, 'INF', 3.0]])
        """
        self.opt_params.set(**kwargs)

        # update kim internal model param (note, set_one will update model_params)
        for name, _ in kwargs.items():
            p_idx = self.model_params[name].index
            for c_idx, v in enumerate(self.model_params[name].value):
                try:
                    self.kim_model.set_parameter(p_idx, c_idx, v)
                except RuntimeError:
                    raise kimpy.KimPyError(
                        "Calling `kim_model.set_parameter()` failed."
                    )

        try:
            self.kim_model.clear_then_refresh()
        except RuntimeError:
            raise kimpy.KimPyError("Calling `kim_model.clear_then_refresh()` failed.")

        # reset influence distance in case it changes
        self.init_influence_distance()

    def set_one_opt_param(self, name: str, settings: List[List[Any]]):
        """
        Set one parameter that will be optimized.

        The name of the parameter should be given as the first entry of a list (or tuple),
        and then each data line should be given in in a list.

        Args:
            name: name of a fitting parameter
            settings: initial value, flag to fix a parameter, lower and upper bounds of a
                parameter.

        Example:
            name = 'param_A'
            settings = [['default', 0, 20], [2.0, 'fix'], [2.2, 'inf', 3.3]]
            instance.set_one(name, settings)
        """
        self.opt_params.set_one(name, settings)

        # update kim internal model param (note, set_one will update model_params)
        p_idx = self.model_params[name].index
        for c_idx, v in enumerate(self.model_params[name].value):
            try:
                self.kim_model.set_parameter(p_idx, c_idx, v)
            except RuntimeError:
                raise kimpy.KimPyError("Calling `kim_model.set_parameter()` failed.")

        try:
            self.kim_model.clear_then_refresh()
        except RuntimeError:
            raise kimpy.KimPyError("Calling `kim_model.clear_then_refresh()` failed.")

        # reset influence distance in case it changes
        self.init_influence_distance()

    def update_model_params(self, params: Sequence[float]):
        """
        Update optimizing parameters (a sequence used by the optimizer) to the kim model.
        """

        # update from opt params to model params
        super().update_model_params(params)

        # only update optimizing params
        if self.params_transform is None:
            # update from model params to kim params
            n = self.get_num_opt_params()
            for i in range(n):
                _, value, p_idx, c_idx = self.get_opt_param_name_value_and_indices(i)
                try:
                    self.kim_model.set_parameter(p_idx, c_idx, value)
                except RuntimeError:
                    raise kimpy.KimPyError(
                        "Calling `kim_model.set_parameter()` failed."
                    )

        # When params_transform is set, a user can do whatever in it
        # function, e.g. update a parameter that is not an optimizing parameter.
        # In general, we do not know how parameters are modified in there,
        # and therefore, we need to update all params in model_params to kim
        # Note, `params_transform.inverse_transform()` is called in
        # super().update_model_params(params)
        else:
            for name, params in self.model_params.items():
                p_idx = params.index
                for c_idx, value in enumerate(params.value):
                    try:
                        self.kim_model.set_parameter(p_idx, c_idx, value)
                    except RuntimeError:
                        raise kimpy.KimPyError(
                            "Calling `kim_model.set_parameter()` failed."
                        )

        # refresh model
        self.kim_model.clear_then_refresh()

        if get_log_level() == "DEBUG":
            params = self.get_kim_model_params()
            s = ""
            for name, p in params.items():
                s += f"\nname: {name}\n"
                s += str(p.as_dict())

            logger.debug(s)

    def write_kim_model(self, path: Path = None):
        """
        Write out a KIM model that can be used directly with the kim-api.

        This function typically write two files to `path`: (1) CMakeLists.txt, and (2)
        a parameter file like A.model_params. `path` will be created if it does not exist.

        Args:
            path: Path to the a directory to store the model. If `None`, it is set to
                `./MODEL_NAME_kliff_trained`, where `MODEL_NAME` is the `model_name` that
                provided at the initialization of this class.

        Note:
            This only works for parameterized KIMModel models that support the writing of
            parameters.
        """
        present, required = self.kim_model.is_routine_present(
            kimpy.model_routine_name.WriteParameterizedModel
        )
        if not present:
            raise KIMModelError("This KIM model does not support writing parameters.")

        if path is None:
            model_name = self.model_name + "_kliff_trained"
            path = Path.cwd().joinpath(model_name)
        else:
            path = Path(path).expanduser().resolve()
            model_name = path.name

        if not path.exists():
            os.makedirs(path)

        path = str(path)
        model_name = str(model_name)

        self.kim_model.write_parameterized_model(path, model_name)

        logger.info(f"KLIFF trained model write to `{path}`")


class KIMModelError(Exception):
    def __init__(self, msg):
        super(KIMModelError, self).__init__(msg)
        self.msg = msg



import numpy as np

from kliff.models.parameter import Parameter
from kliff.models.parameter_transform import LogParameterTransform


def test_log_transform():

    a_val = [1.0, 2.0]
    b_val = [3.0, 4.0]

    a = Parameter(a_val)
    b = Parameter(b_val)
    params = {"a": a, "b": b}

    transformer = LogParameterTransform(param_names=["b"])

    forward = transformer.transform(params)
    assert np.allclose(forward["a"].value, a_val)
    assert np.allclose(forward["b"].value, np.log(b_val))

    inverse = transformer.inverse_transform(forward)
    assert np.allclose(inverse["a"].value, a_val)
    assert np.allclose(inverse["b"].value, b_val)



from typing import Optional

import numpy as np

from kliff.error import report_import_error

try:
    import emcee

    emcee_avail = True
except ImportError:
    emcee_avail = False


# Estimate the burn-in time
def mser(
    chain: np.ndarray,
    dmin: Optional[int] = 1,
    dstep: Optional[int] = 10,
    dmax: Optional[int] = -1,
    full_output: Optional[bool] = False,
) -> int:
    """Estimate the equilibration time using marginal standard error rule (MSER). This is
    done by calculating the standard error (square) of ``chain_d``, where ``chain_d``
    contains the last :math:`n-d` element of the chain (n is the total number of
    iterations for each chain), for progresively larger d values, starting from ``dmin``
    upto ``dmax``, incremented by ``dstep``. The SE values are stored in a list. Then we
    search the minimum element in the list and return the index of that element.

    Parameters
    ----------
    chain: 1D np.ndarray
        Array containing the time series.
    dmin: int
        Index where to start the search in the time series.
    dstep: int
        How much to increment the search is done.
    dmax: int
        Index where to stop the search in the time series.
    full_output: bool
        A flag to return the list of squared standard error.

    Returns
    -------
    dstar: int or dict
        Estimate of the equilibration time using MSER. If ``full_output=True``, then a
        dictionary containing the estimated equilibration time and the list of squared
        standard errors will be returned.
    """
    length = len(chain)  # Chain length

    # Compute the SE square
    SE2_list = [
        _standard_error_squared(chain[dd:]) for dd in range(length)[dmin:dmax:dstep]
    ]

    # Get the estimate of the equilibration time, wrt the original time series
    dest = np.argmin(SE2_list)
    dstar = min([dmin + (dest + 1) * dstep, length])

    if full_output:
        return {"dstar": dstar, "SE2": SE2_list}
    else:
        return dstar


# Estimate autocorrelation length
def autocorr(chain: np.ndarray, *args, **kwargs):
    """Use ``emcee`` package to estimate the autocorrelation length.

    Parameters
    ----------
    chain: np.ndarray (nwalkers, nsteps, ndim,)
        Chains from the MCMC simulation. The shape of the chains needs to be
        (nsteps, nwalkers, ndim). Note that the burn-in time needs to be discarded prior
        to this calculation
    args, kwargs
        Additional positional and keyword arguments of ``emcee.autocorr.integrated_time``.

    Returns
    -------
    float or array:
        Estimate of the autocorrelation length for each parameter.
    """
    if emcee_avail:
        chain = np.swapaxes(chain, 0, 1)
        return emcee.autocorr.integrated_time(chain, *args, **kwargs)
    else:
        report_import_error("emcee")


# Assess convergence
def rhat(chain, time_axis: int = 1, return_WB: bool = False):
    """Compute the value of :math:`\\hat{r}` proposed by Brooks and Gelman
    [BrooksGelman1998]_. If the samples come from PTMCMC simulation, then the chain needs
    to be from one of the temperature only.

    Parameters
    ----------
    chain: ndarray
        The MCMC chain as a ndarray, preferrably with the shape
        (nwalkers, nsteps, ndims). However, the shape can also be
        (nsteps, nwalkers, ndims), but the argument time_axis needs to be set
        to 0.
    time_axis: int (optional)
        Axis in which the time series is stored (0 or 1). For emcee results,
        the time series is stored in axis 0, but for ptemcee for a given
        temperature, the time axis is 1.
    return_WB: bool (optional)
        A flag to return covariance matrices within and between chains.

    Returns
    -------
    r: float
        The value of rhat.
    W, B: 2d ndarray
        Matrices of covariance within and between the chains.

    References
    ----------
    .. [BrooksGelman1998]
       Brooks, S.P., Gelman, A., 1998. General Methods for Monitoring Convergence of
       Iterative Simulations. Journal of Computational and Graphical Statistics 7,
       434455. https://doi.org/10.1080/10618600.1998.10474787
    """
    if time_axis == 1:
        # Reshape the chain so that the time axis is in axis 1
        temp = np.swapaxes(chain, 0, 1)
        chain = temp

    m, n, _ = chain.shape
    lambda1, W, B = _lambda1(chain)
    r = 1 - 1 / n + (1 + 1 / m) * lambda1

    if return_WB:
        toreturn = (r, W, B)
    else:
        toreturn = r

    return toreturn


def _standard_error_squared(chain: np.ndarray) -> float:
    """Compute the square of the standard error."""
    nn = len(chain)
    se2 = np.var(chain) / nn
    return se2


def _lambda1(chain):
    """Compute the largest eigenvalue of :math:`W^{-1} B/n`."""
    W = _W(chain)
    B_over_n = _B_over_n(chain)
    V = np.linalg.lstsq(W, B_over_n, rcond=-1)[0]
    s = np.linalg.svd(V, compute_uv=False)
    return np.max(s), W, B_over_n


def _B_over_n(chain):
    """Compute covariance matrix between the chains."""
    return np.cov(np.mean(chain, axis=1), rowvar=False, ddof=1)


def _W(chain):
    """Compute the mean of the covariance matrix within each chain."""
    m, n, nparams = chain.shape
    Wm = np.empty((m, nparams, nparams))
    for walker in range(m):
        Wm[walker] = np.cov((chain[walker]), rowvar=False, ddof=1)
    return np.mean(Wm, axis=0)



"""
.. _tut_params_transform:

Parameter transformation for the Stillinger-Weber potential
===========================================================

Parameters in the empirical interatomic potential are often restricted by some physical
constraints. As an example, in the Stillinger-Weber (SW) potential, the energy scaling
parameters (e.g., ``A`` and ``B``) and the length scaling parameters (e.g., ``sigma`` and
``gamma``) are constrained to be positive.

Due to these constraints, we might want to work with the log of the parameters, i.e.,
``log(A)``, ``log(B)``, ``log(sigma)``, and ``log(gamma)`` when doing the optimization.
After the optimization, we can transform them back to the original parameter space using
an exponential function, which will guarantee the positiveness of the parameters.

In this tutorial, we show how to apply parameter transformation to the SW potential for
silicon that is archived on OpenKIM_. Compare this with :ref:`tut_kim_sw`.
"""


##########################################################################################
# To start, let's first install the SW model::
#
#    $ kim-api-collections-management install user SW_StillingerWeber_1985_Si__MO_405512056662_006
#
# .. seealso::
#    This installs the model and its driver into the ``User Collection``. See
#    :ref:`install_model` for more information about installing KIM models.
#
# This is

import numpy as np

from kliff.calculators import Calculator
from kliff.dataset import Dataset
from kliff.dataset.weight import Weight
from kliff.loss import Loss
from kliff.models import KIMModel
from kliff.models.parameter_transform import LogParameterTransform
from kliff.utils import download_dataset

##########################################################################################
# Before creating a KIM model for the SW potential, we first instantiate the parameter
# transformation class that we want to use. ``kliff`` has a built-in log-transformation;
# however, extending it to other parameter transformation can be done by creating a
# subclass of :class:`~kliff.models.parameter_transform.ParameterTransform`.
#
# To make a direct comparison to :ref:`tut_kim_sw`, in this tutorial we will apply
# log-transformation to parameters ``A``, ``B``, ``sigma``, and ``gamma``, which
# correspond to energy and length scales.
#

transform = LogParameterTransform(param_names=["A", "B", "sigma", "gamma"])
model = KIMModel(
    model_name="SW_StillingerWeber_1985_Si__MO_405512056662_006",
    params_transform=transform,
)
model.echo_model_params(params_space="original")


##########################################################################################
# ``model.echo_model_params(params_space="original")`` above will print out parameter
# values in the original, untransformed space, i.e., the original parameterization of
# the model. If we supply the argument ``params_space="transformed"``, then the printed
# parameter values are given in the transformed space, e.g., log space (below). The
# values of the other parameters are not changed.
#

model.echo_model_params(params_space="original")


##########################################################################################
# Compare the output of ``params_space="transformed"`` and ``params_space="original"``,
# you can see that the values of ``A``, ``B``, ``sigma``, and ``gamma`` are in the log
# space after the transformation.

##########################################################################################
# Next, we will set up the initial guess of the parameters to optimize. A value of
# ``"default"`` means the initial guess will be directly taken from the value already in
# the model.
#
# .. note::
#    The parameter values we initialize, as well as the lower and upper bounds, are in
#    transformed space (i.e. log space here).


model.set_opt_params(
    A=[[np.log(5.0), np.log(1.0), np.log(20)]],
    B=[["default"]],
    sigma=[[np.log(2.0951), "fix"]],
    gamma=[[np.log(1.5)]],
)
model.echo_opt_params()

##########################################################################################
# We can show the parameters we’ve just set by ``model.echo_opt_params()``.
#
# .. note::
#    ``model.echo_opt_params()`` always displays the parameter values in the transformed
#    space. And it only shows all the parameters specified to optimize. To show all
#    the parameters, do ``model.echo_model_params(params_space="transformed")``.

##########################################################################################
# Once we set the model and the parameter transformation scheme, then further
# calculations, e.g., training the model, will be performed using the transformed space
# and can be done in the same way as in :ref:`tut_kim_sw`.

# Training set
dataset_path = download_dataset(dataset_name="Si_training_set")
weight = Weight(energy_weight=1.0, forces_weight=0.1)
tset = Dataset(dataset_path, weight)
configs = tset.get_configs()

# Calculator
calc = Calculator(model)
_ = calc.create(configs)

# Loss function and model training
steps = 100
loss = Loss(calc, nprocs=2)
loss.minimize(method="L-BFGS-B", options={"disp": True, "maxiter": steps})

model.echo_model_params(params_space="original")


##########################################################################################
# The optimized parameter values from this model training are very close, if not the
# same, as in :ref:`tut_kim_sw`. This is expected for the simple tutorial example
# considered. But for more complex models, training in a transformed space can make it
# much easier for the optimizer to navigate the parameter space.
#
# .. _OpenKIM: https://openkim.org
#



import warnings
from typing import Callable, List, Optional

import numpy as np

from kliff.error import report_import_error
from kliff.loss import Loss

try:
    import emcee

    emcee_avail = True
except ImportError:
    emcee_avail = False

try:
    import ptemcee

    ptemcee_avail = True
except ImportError:
    ptemcee_avail = False


def logprior_uniform(x: np.ndarray, bounds: np.ndarray) -> float:
    """Logarithm of the non-normalized joint uniform prior. This is the default prior
    distribution.

    Parameters
    ----------
    x : np.ndarray
        Parameter values to evaluate.
    bounds : np.ndarray (nparams, 2,)
        An array containing the boundaries of the uniform prior. The first column of the
        array contains the lower bounds and the second column contains the upper bounds.

    Returns
    -------
    float:
        Logarithm of the non-normalized joint uniform prior evaluated at parameter ``x``.
    """
    l_bounds, u_bounds = bounds.T
    if all(np.less(x, u_bounds)) and all(np.greater(x, l_bounds)):
        ret = 0.0
    else:
        ret = -np.inf

    return ret


def get_T0(loss: Loss):
    """Compute the natural temperature. The minimum loss is the loss value at the optimal
    parameters.

    Parameters
    ----------
    loss : Loss
        Loss function class from :class:`~kliff.loss.Loss`.

    Returns
    -------
    float
        Value of the natural temperature.
    """
    xopt = loss.calculator.get_opt_params()
    ndim = len(xopt)
    C0 = loss._get_loss(xopt)
    return 2 * C0 / ndim


class MCMC:
    """MCMC sampler class for Bayesian uncertainty quantification.

    This is a wrapper over :class:`PtemceeSampler` and
    :class:`EmceeSampler`. Currently, only these  2 samplers implemented.

    Parameters
    ----------
    loss : Loss
        Loss function class from :class:`~kliff.loss.Loss`.
    nwalkers : Optional[int]
        Number of walkers to simulate. The minimum number of walkers
        is twice the number of parameters. It defaults to this minimum
        value.
    logprior_fn : Optional[Callable]
        A function that evaluate logarithm of the prior
        distribution. The prior doesn't need to be normalized. It
        defaults to a uniform prior over a finite range.
    logprior_args : Optional[tuple]
        Additional positional arguments of the ``logprior_fn``. If the
        default ``logprior_fn`` is used, then the boundaries of the
        uniform prior can be specified here.
    sampler : Optional[str] or sampler instance
        An argument that specifies the MCMC sampler to use. The value can be one of the
        strings ``"ptemcee"`` (the default value) or ``"emcee"``, or a sampler class
        instance. If ``"ptemcee"`` or ``"emcee"`` is given, a respective internal sampler
        class will be uses.
    **kwargs : Optional[dict]
        Additional keyword arguments for ``ptemcee.Sampler`` or ``emcee.EnsembleSampler``.
    """

    builtin_samplers = ["ptemcee", "emcee"]

    def __new__(
        self,
        loss: Loss,
        nwalkers: Optional[int] = None,
        logprior_fn: Optional[Callable] = None,
        logprior_args: Optional[tuple] = None,
        sampler: Optional[str] = "ptemcee",
        **kwargs,
    ):

        if sampler == "ptemcee":
            if ptemcee_avail:
                # Force pool to be declared after MCMC is instantiated
                if "pool" in kwargs:
                    raise ValueError(
                        "Please declare the pool after instantiating ``kliff.uq.MCMC``"
                    )

                # Pre-process keyword arguments
                ntemps = kwargs.pop("ntemps") if "ntemps" in kwargs else 10
                Tmax_ratio = kwargs.pop("Tmax_ratio") if "Tmax_ratio" in kwargs else 1.0
                Tladder = kwargs.pop("Tladder") if "Tladder" in kwargs else None

                return PtemceeSampler(
                    loss,
                    nwalkers,
                    ntemps,
                    Tmax_ratio,
                    Tladder,
                    logprior_fn,
                    logprior_args,
                    **kwargs,
                )
            else:
                report_import_error("ptemcee")
        elif sampler == "emcee":
            if emcee_avail:
                # Force pool to be declared after MCMC is instantiated
                if "pool" in kwargs:
                    raise ValueError(
                        "Please declare the pool after instantiating ``kliff.uq.MCMC``"
                    )

                # Pre-process keyword arguments
                T = kwargs.pop("T") if "T" in kwargs else None

                return EmceeSampler(
                    loss, nwalkers, T, logprior_fn, logprior_args, **kwargs
                )
            else:
                report_import_error("emcee")
        elif isinstance(sampler, str) and sampler not in self.builtin_samplers:
            raise ValueError(
                "Unknown sampler. Please input 'ptemcee', 'emcee' or "
                "some sampler class instance"
            )
        else:
            warnings.warn("The other arguments are ignored")
            return sampler


if ptemcee_avail:

    class PtemceeSampler(ptemcee.Sampler):
        """Sampler class for PTMCMC via ``ptemcee`` Python package.

        Parameters
        ----------
        loss : Loss
            Loss function instance from :class:`~kliff.loss.Loss`.
        nwalkers : Optional[int]
            Number of walkers to simulate. The minimum number of walkers
            is twice the number of parameters. It defaults to this minimum
            value.
        ntemps: Optional[int]
            Number of temperatures to simulate. It defaults to 10.
        Tmax_ratio: Optional[float]
            The ratio between the highest temperature to use and the natural temperature.
            Higher value means that the maximum temperature is higher than :math:`T_0`
            [Frederiksen2004]_. It defaults to 1.0.
        Tladder: Optional[List]
            A list containing the temperature values to use. The values nedd to be
            monotonically increasing or decreasing. It defaults to ``None``, which will
            be generated from ``ntemps`` and ``Tmax_ratio``.
        logprior_fn : Optional[Callable]
            A function that evaluate logarithm of the prior
            distribution. The prior doesn't need to be normalized. It
            defaults to a uniform prior over a finite range.
        logprior_args : Optional[tuple]
            Additional positional arguments of the ``logprior_fn``. If the
            default ``logprior_fn`` is used, then the boundaries of the
            uniform prior can be specified here.
        **kwargs : Optional[dict]
            Additional keyword arguments for ``ptemcee.Sampler``.

        Attributes
        ----------
        loss: Loss
            Loss function instance from :class:`~kliff.loss.Loss`
        T0: float
            Values of the natural temperature, :math:`T_0` [Frederiksen2004]_.
        Tladder: np.ndarray
            An array containing the values of the sampling temperatures.
        sampler: ptemcee.Sampler
            Sampler instance from ``ptemcee.Sampler``
        """

        def __init__(
            self,
            loss: Loss,
            nwalkers: Optional[int] = None,
            ntemps: Optional[int] = 10,
            Tmax_ratio: Optional[float] = 1.0,
            Tladder: Optional[List] = None,
            logprior_fn: Optional[Callable] = None,
            logprior_args: Optional[tuple] = None,
            **kwargs,
        ):
            self.loss = loss

            # Dimensionality
            ndim = loss.calculator.get_num_opt_params()
            nwalkers = 2 * ndim if nwalkers is None else nwalkers

            # Probability
            global loglikelihood_fn

            def loglikelihood_fn(x):
                return _get_loglikelihood(x, loss)

            if logprior_fn is None:
                logprior_fn = logprior_uniform
                if logprior_args is None:
                    logprior_args = (_get_parameter_bounds(loss),)

            # Sampling temperatures
            self.T0 = None
            self.Tladder = self._generate_temp_ladder(ntemps, Tmax_ratio, Tladder)
            betas = 1 / self.Tladder

            super().__init__(
                nwalkers,
                ndim,
                loglikelihood_fn,
                logprior_fn,
                logpargs=logprior_args,
                betas=betas,
                **kwargs,
            )

        def _generate_temp_ladder(self, ntemps, Tmax_ratio, Tladder):
            """Generate temperature ladder"""
            # Only generate temperature ladder when it is not specified.
            if Tladder is None:
                # Compute T0
                self.T0 = get_T0(self.loss)

                Tmax = Tmax_ratio * self.T0
                Tmax_not_T0 = Tmax_ratio != 1.0

                # If Tmax is not T0, then we need to generate 1 less temperature than
                # requested, and append T0 afterward.
                if Tmax_not_T0:
                    ntemps -= 1

                Tladder = np.logspace(0, np.log10(Tmax), ntemps)

                if Tmax_not_T0:
                    Tladder = np.sort(np.append(Tladder, self.T0))
            return Tladder


if emcee_avail:

    class EmceeSampler(emcee.EnsembleSampler):
        """Sampler class for affine invariant MCMC via ``emcee`` Python package.

        Parameters
        ----------
        loss : Loss
            Loss function instance from :class:`~kliff.loss.Loss`.
        nwalkers : Optional[int]
            Number of walkers to simulate. The minimum number of walkers
            is twice the number of parameters. It defaults to this minimum
            value.
        T: Optional[float]
            Sampling temperatures, used to inflate the likelihood function in the MCMC
            sampling. It defaults to the natural temperature :math:`T_0` [Frederiksen2004]_.
        logprior_fn : Optional[Callable]
            A function that evaluate logarithm of the prior
            distribution. The prior doesn't need to be normalized. It
            defaults to a uniform prior over a finite range.
        logprior_args : Optional[tuple]
            Additional positional arguments of the ``logprior_fn``. If the
            default ``logprior_fn`` is used, then the boundaries of the
            uniform prior can be specified here.
        **kwargs : Optional[dict]
            Additional keyword arguments for ``emcee.EnsembleSampler``.

        Attributes
        ----------
        loss: Loss
            Loss function instance from :class:`~kliff.loss.Loss`
        T: float
            Values of the sampling temperature.
        sampler: emcee.EnsembleSampler
            Sampler instance from ``emcee.EnsembleSampler``

        Notes
        -----
        As a convention, KLIFF inflates the likelihood by some sampling temperature, i.e.,
        :math:`L(\\theta) \propto \exp(-C(\\theta) / T)`. As a default, the sampling
        temperature is set to the natural temperature. To use the untempered likelihood
        (:math:`T=1`), user should specify the argument ``T=1``.


        References
        ----------
        .. [Frederiksen2004] S. L. Frederiksen, K. W. Jacobsen, K. S. Brown, and J. P.
           Sethna, “Bayesian Ensemble Approach to Error Estimation of Interatomic
           Potentials,” Phys. Rev. Lett., vol. 93, no. 16, p. 165501, Oct. 2004,
           doi: 10.1103/PhysRevLett.93.165501.
        """

        def __init__(
            self,
            loss: Loss,
            nwalkers: Optional[int] = None,
            T: Optional[float] = None,
            logprior_fn: Optional[Callable] = None,
            logprior_args: Optional[tuple] = None,
            **kwargs,
        ):
            self.loss = loss

            # Dimensionality
            ndim = loss.calculator.get_num_opt_params()
            nwalkers = 2 * ndim if nwalkers is None else nwalkers

            # Probability
            if T is None:
                self.T = get_T0(self.loss)
            else:
                self.T = T
            logl_fn = self._loglikelihood_wrapper
            logp_fn = self._logprior_wrapper(logprior_fn, *logprior_args)

            global logprobability_fn

            def logprobability_fn(x):
                return logl_fn(x) + logp_fn(x)

            super().__init__(nwalkers, ndim, logprobability_fn, **kwargs)

        def _loglikelihood_wrapper(self, x):
            """A wrapper to the log-likelihood function, so that the only argument is the
            parameter values.
            """
            return _get_loglikelihood(x, self.loss, self.T)

        def _logprior_wrapper(self, logprior_fn, *logprior_args):
            """A wapper to the log-prior function, so that the only argument is the
            parameter values.
            """
            if logprior_fn is None:
                if logprior_args is None:
                    logprior_args = (_get_parameter_bounds(self.loss),)
                logprior_fn = logprior_uniform

            def logp(x):
                return logprior_fn(x, *logprior_args)

            return logp


def _get_loglikelihood(x: np.ndarray, loss: Loss, T: Optional[float] = 1.0):
    """Compute the log-likelihood from the cost function. It has an option to temper the
    cost by specifying ``T``.
    """
    cost = loss._get_loss(x)
    logl = -cost / T
    return logl


def _get_parameter_bounds(loss):
    """Get the parameter bounds for the default uniform prior."""
    bounds = loss.calculator.get_opt_params_bounds()
    if np.any(np.isin(bounds, None)):
        raise MCMCError(
            "Improper uniform prior bound is used. Please specify a finite range."
        )
    return np.array(bounds)


class MCMCError(Exception):
    def __init__(self, msg):
        super(MCMCError, self).__init__(msg)
        self.msg = msg



from collections.abc import Sequence
from typing import Any, Dict, List, Union

import numpy as np
from loguru import logger

from kliff.dataset.dataset import Configuration
from kliff.models.model import Model
from kliff.utils import length_equal


class Calculator:
    """
    Calculator to compute properties (e.g. energy and forces) using a potential model for
    each configuration in the dataset.

    The properties, together with the corresponding reference data stored in the
    configurations, are provided to :class:`~kliff.loss.Loss` to construct a loss
    function for the optimizer.

    In the reverse direction, a calculator grabs the new parameters from the optimizer
    and update the model with the new parameters.

    Args:
        model: An instance of :class:`~kliff.models.Model`.
    """

    def __init__(self, model: Model):
        self.model = model

        self.compute_arguments = None

    def create(
        self,
        configs: List[Configuration],
        use_energy: Union[List[bool], bool] = True,
        use_forces: Union[List[bool], bool] = True,
        use_stress: Union[List[bool], bool] = False,
    ):
        """
        Create compute arguments for a collection of configurations.

        By compute arguments, we mean the information needed by a model to carry on a
        calculation, such as the coordinates, species, cutoff distance, neighbor list,
        etc. Each configuration has its own compute arguments, and this function creates
        the compute arguments for all the configurations in ``configs``.

        Args:
            configs: atomic configurations.
                instances.
            use_energy: Whether to require the calculator to compute energy.
                If a list of bool is provided, each component is for one configuration
                in `configs`. If a bool is provided, it is applied to all configurations.
            use_forces: Whether to require the calculator to compute forces.
                If a list of bool is provided, each component is for one configuration
                in `configs`. If a bool is provided, it is applied to all configurations.
            use_stress: Whether to require the calculator to compute stress.
                If a list of bool is provided, each component is for one configuration
                in `configs`. If a bool is provided, it is applied to all configurations.
        """

        self.use_energy = use_energy
        self.use_forces = use_forces
        self.use_stress = use_stress

        if isinstance(configs, Configuration):
            configs = [configs]

        if use_energy and not length_equal(configs, use_energy):
            raise CalculatorError(
                "Expect length of `configs` and `use_energy` have the same size; got "
                f"{len(configs)} and {len(use_energy)}."
            )
        if use_forces and not length_equal(configs, use_forces):
            raise CalculatorError(
                "Expect length of `configs` and `use_forces` have the same size; got "
                f"{len(configs)} and {len(use_forces)}."
            )
        if use_stress and not length_equal(configs, use_stress):
            raise CalculatorError(
                "Expect length of `configs` and `use_stresses` have the same size; got "
                f"{len(configs)} and {len(use_stress)}."
            )

        N = len(configs)
        if not isinstance(use_energy, Sequence):
            use_energy = [use_energy] * N
        if not isinstance(use_forces, Sequence):
            use_forces = [use_forces] * N
        if not isinstance(use_stress, Sequence):
            use_stress = [use_stress] * N

        supported_species = self.model.get_supported_species()
        infl_dist = self.model.get_influence_distance()

        ca_class = self.model.get_compute_argument_class()

        self.compute_arguments = []
        for conf, e, f, s in zip(configs, use_energy, use_forces, use_stress):
            if self._is_kim_model():
                kim_ca = self.model.create_a_kim_compute_argument()
                ca = ca_class(kim_ca, conf, supported_species, infl_dist, e, f, s)
            else:
                ca = ca_class(conf, supported_species, infl_dist, e, f, s)
            self.compute_arguments.append(ca)

        logger.info(f"Create calculator for {len(configs)} configurations.")
        return self.compute_arguments

    def get_compute_arguments(self) -> List[Any]:
        """
        Return a list of compute arguments, each associated with a configuration.
        """
        return self.compute_arguments

    # TODO, maybe change the compute_argument.compute api of kim models, such
    #   that it accept params as the argument, instead of kim_model
    def compute(self, compute_arguments) -> Dict[str, Any]:
        """
        Compute the properties given the compute arguments associated with a
        configuration.

        Args:
            compute_arguments: A compute arguments instance for a configuration.

        Returns:
            A dictionary of properties, with keys of `energy`, `forces` and `stress`,
                values of float or np.array.
        """
        if self._is_kim_model():
            compute_arguments.compute(self.model.kim_model)
        else:
            compute_arguments.compute(self.model.get_model_params())
        return compute_arguments.results

    def get_energy(self, compute_arguments) -> float:
        """
        Get the energy of a configuration.

        Args:
            compute_arguments: A compute arguments instance for a configuration.

        Returns:
            The energy of the configuration associated with the compute arguments.
        """
        return compute_arguments.get_energy()

    def get_forces(self, compute_arguments) -> np.array:
        """
        Get the forces of a configuration.

        Args:
            compute_arguments: A compute arguments instance for a configuration.

        Returns:
            Forces on atoms. 2D array of shape (N, 3), where N is the number of atoms
                in the configuration.
        """
        return compute_arguments.get_forces()

    def get_stress(self, compute_arguments) -> np.array:
        r"""
        Get the stress of a configuration.

        Args:
            compute_arguments: A compute arguments instance for a configuration.

        Returns:
            Virial stress of the configuration associated with the compute arguments,
            1D array with 6 component.
            The returned stress is in Voigt notation, i.e. this function returns:
            [:math:`\sigma_{xx}, \sigma_{yy}, \sigma_{zz}, \sigma_{yz}, \sigma_{xy},
            \sigma_{xz}`]
        """
        return compute_arguments.get_stress()

    # NOTE, we keep the interface of get_prediction, get_opt_params, has_opt_params
    # etc... to make it easier to implement WrapperCalculators, where multiple
    # calculators are used.
    def get_prediction(self, compute_arguments) -> np.array:
        """
        Get the prediction of all properties that are requested to compute.

        The `energy`, `forces`, and `stress` are each flattened to a 1D array, and then
        concatenated (in the order of `energy`, `forces`, and `stress`) to form the
        prediction.  Depending on the values of ``use_energy``, ``use_forces``, and
        ``use_stress`` that are provided in :meth:`create`, one or more of `energy`,
        `forces` and `stress` may not be included in the prediction.

        Args:
            compute_arguments: A compute arguments instance for a configuration.

        Returns:
            1D array of predictions. For a configuration of N atoms:
            - If ``use_energy``, ``use_forces``, and ``use_stress`` are all ``True``, the
              size of prediction is `1+3N+6`, with the 1st component the `energy`, the 2nd
              to 1+3N components the flattened `forces` on N atoms, and the remaining 6
              components the Voigt `stress`.
            - If one or more of ``use_energy``, ``use_forces``, and ``use_stress`` is
              ``False``, its corresponding value is removed from prediction, and the size
              of prediction shrinks accordingly. For example, if both ``use_energy`` and
              ``use_stress`` are ``True`` but ``use_forces`` is ``False``, then the size
              of prediction is `1+6`, with the 1st component the `energy`, and the 2nd to
              7th components the Voigt stress.
        """
        return compute_arguments.get_prediction()

    def get_reference(self, compute_arguments) -> np.array:
        """
        Get the reference data of all properties that are requested to compute.

        Same as :meth:`get_prediction`, the `energy`, `forces`, and `stress` are each
        flattened to a 1D array, and then concatenated (in the order of `energy`,
        `forces`, and `stress`) to form the reference. Depending on the values of
        ``use_energy``, ``use_forces``, and ``use_stress`` that are provided in
        :meth:`create`, one or more of `energy`, `forces` and `stress` may not be included
        in the reference.

        Args:
            compute_arguments: A compute arguments instance for a configuration.

        Returns:
            1D array of reference values. For a configuration of N atoms:
            - If ``use_energy``, ``use_forces``, and ``use_stress`` are all ``True``, the
              size of reference is `1+3N+6`, with the 1st component the `energy`, the 2nd
              to 1+3N components the flattened `forces` on N atoms, and the remaining 6
              components the Voigt `stress`.
            - If one or more of ``use_energy``, ``use_forces``, and ``use_stress`` is
              ``False``, its corresponding value is removed from reference, and the size
              of reference shrinks accordingly. For example, if both ``use_energy`` and
              ``use_stress`` are ``True`` but ``use_forces`` is ``False``, then the size
              of reference is `1+6`, with the 1st component the `energy`, and the 2nd to
              7th components the Voigt stress.
        """
        return compute_arguments.get_reference()

    def get_opt_params(self) -> np.array:
        """
        Return a list of optimizing parameters.

        The optimizing parameters is a list consisting of the values of the model
        parameters that is set to fit via :meth:`kliff.models.Model.set_opt_params` or
        :meth:`kliff.models.Model.set_one_opt_param`.  The returned value is typically
        passed to an optimizer as the initial values to carry out the optimization.
        """
        return self.model.get_opt_params()

    def get_num_opt_params(self) -> int:
        """
        Return the number of optimizing parameters.
        """
        return self.model.get_num_opt_params()

    def has_opt_params_bounds(self):
        """
        Return a bool to indicate whether there are parameters whose bounds are
        provided.
        """
        return self.model.has_opt_params_bounds()

    def get_opt_params_bounds(self):
        """
        Return the lower and upper bounds for the optimizing parameters.

        The returned value is a list of (lower, upper) tuples. Each tuple contains the
        lower and upper bounds for the corresponding parameter obtained from
        :meth:`get_opt_params`. ``None`` for ``lower`` or ``upper`` means that no bound
        should be applied.
        """
        return self.model.get_opt_params_bounds()

    def update_model_params(self, opt_params: List[float]):
        """
        Update model parameters from a list of opt_params.

        This is the reverse of :meth:`get_opt_params`.

        Args:
            opt_params: Optimizing parameters.
        """
        self.model.update_model_params(opt_params)

    def _is_kim_model(self):
        return self.model.__class__.__name__ == "KIMModel"


class _WrapperCalculator(object):
    """
    Wrapper to deal with the fitting of multiple models.

    Args:
        calculators: calculators to wrap. Each calculator for a different model,
        but should be for the same dataset.
    """

    # TODO methods like get_prediction needs to be implemented. The prediction of
    #  energy (force) should be the sum of each calculator.
    def __init__(self, calculators=List[Calculator]):
        """
        Parameters
        ----------

        calculators: instance of Calculator
        """
        self.calculators = calculators
        self._start_end = self._set_start_end()

    def _set_start_end(self):
        """
        Compute the start and end indices of the `opt_params` of each calculator in the
        `opt_params` of the wrapper calculator.
        """
        start_end = []
        i = 0
        for calc in self.calculators:
            n = calc.get_num_opt_params()
            start = i
            end = i + n
            start_end.append((start, end))
            i += n
        return start_end

    def get_compute_arguments(self):
        all_cas = []
        for calc in self.calculators:
            cas = calc.get_compute_arguments()
            all_cas.extend(cas)
        return all_cas

    def get_num_opt_params(self):
        return sum([calc.get_num_opt_params() for calc in self.calculators])

    def get_opt_params(self):
        return np.concatenate([calc.get_opt_params() for calc in self.calculators])

    def get_opt_params_bounds(self):
        bounds = []
        for calc in self.calculators:
            b = calc.get_opt_params_bounds()
            bounds.extend(b)
        return bounds

    def update_model_params(self, opt_params):
        for i, calc in enumerate(self.calculators):
            start, end = self._start_end[i]
            p = opt_params[start:end]
            calc.update_model_params(p)

    def get_calculator_list(self):
        """
        Create a list of calculators.

        Each calculator has `number of configurations` copies in the list.

        In such, we can easily associate the compute arguments with the calculator.
        """
        calc_list = []
        for calc in self.calculators:
            N = len(calc.get_compute_arguments())
            calc_list.extend([calc] * N)
        return calc_list


class CalculatorError(Exception):
    def __init__(self, msg):
        super(CalculatorError, self).__init__(msg)
        self.msg = msg



import pickle
import sys
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
from loguru import logger

from kliff import parallel
from kliff.dataset import Configuration
from kliff.utils import create_directory, pickle_dump, to_path


class Descriptor:
    """
    Base class of atomic environment descriptors.

    Process dataset to generate fingerprints. This is the base class for all descriptors,
    so it should not be used directly. Instead, descriptors built on top of this such as
    :class:`~kliff.descriptors.SymmetryFunction` and
    :class:`~kliff.descriptors.Bispectrum` can be used to transform the atomic environment
    information into fingerprints.

    Args:
        cut_dists: Cutoff distances, with key of the form `A-B` where `A` and `B` are
            species string, and value should be a float.
            Example: `cut_dists = {'C-C': 5.0}`
        cut_name: Name of the cutoff function, such as `cos`, `P3`, and `P7`.
        hyperparams: A dictionary of the hyperparams of the descriptor or a string to
            select the predefined hyperparams.
        normalize: If `True`, the fingerprints is centered and normalized:
            `zeta = (zeta - mean(zeta)) / stdev(zeta)`
        dtype: np.dtype
            Data type of the generated fingerprints, such as `np.float32` and `np.float64`.

    Attributes:
        size: int
            Length of the fingerprint vector.
        mean: list
            Mean of the fingerprints.
        stdev: list
            Standard deviation of the fingerprints.
    """

    def __init__(
        self,
        cut_dists: Dict[str, float],
        cut_name: str,
        hyperparams: Union[Dict, str],
        normalize: bool = True,
        dtype=np.float32,
    ):

        self.cut_dists = cut_dists
        self.cut_name = cut_name
        self.hyperparams = hyperparams
        self.normalize = normalize
        self.dtype = dtype

        # size, mean, and stdev of fingerprints; mean and stdev will be used only when
        # `normalize=True`
        self.size = None
        self.mean = None
        self.stdev = None

    def generate_fingerprints(
        self,
        configs: List[Configuration],
        fit_forces: bool = False,
        fit_stress: bool = False,
        fingerprints_filename: Union[Path, str] = "fingerprints.pkl",
        fingerprints_mean_stdev_filename: Optional[Union[Path, str]] = None,
        use_welford_method: bool = False,
        nprocs: int = 1,
    ):
        """
        Convert all configurations to their fingerprints.

        Args:
            configs: Dataset configurations
            fit_forces: Whether to compute the gradient of fingerprints w.r.t. atomic
                coordinates so as to compute forces.
            fit_stress: Whether to compute the gradient of fingerprints w.r.t. atomic
                coordinates so as to compute stress.
            use_welford_method: Whether to compute mean and standard deviation using the
                Welford method, which is memory efficient. See
                https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance
            fingerprints_filename: Path to dump fingerprints to a pickle file.
            fingerprints_mean_stdev_filename: Path to dump the mean and standard
                deviation of the fingerprints as a pickle file. If `normalize=False`
                for the descriptor, this is ignored.

            nprocs: Number of processes used to generate the fingerprints. If `1`, run
                in serial mode, otherwise `nprocs` processes will be forked via
                multiprocessing to do the work.
        """
        if self.mean is not None and self.stdev is not None:
            has_mean_stdev = True
        else:
            has_mean_stdev = False

        # compute mean and stdev
        if self.normalize and not has_mean_stdev:
            logger.info("Start computing mean and stdev of fingerprints.")

            if use_welford_method:
                logger.info(
                    "Computing fingerprint mean and stdev using the Welford method."
                )
                self.mean, self.stdev = self._welford_mean_and_stdev(configs)
                zeta, dzetadr_forces, dzetadr_stress = None, None, None

            else:
                zeta, dzetadr_forces, dzetadr_stress = self._calc_zeta_dzetadr(
                    configs, fit_forces, fit_stress, nprocs
                )
                stacked = np.concatenate(zeta)
                self.mean = np.mean(stacked, axis=0)
                self.stdev = np.std(stacked, axis=0)

            logger.info("Finish computing mean and stdev of fingerprints.")

            # save to a pickle file
            if fingerprints_mean_stdev_filename is None:
                fingerprints_mean_stdev_filename = "fingerprints_mean_and_stdev.pkl"
            state_dict = self.state_dict()
            pickle_dump(state_dict, fingerprints_mean_stdev_filename)

            logger.info(
                "Fingerprints mean and stdev saved to "
                f"`{fingerprints_mean_stdev_filename}`."
            )

        else:
            zeta, dzetadr_forces, dzetadr_stress = None, None, None

        # generate fingerprints
        self._dump_fingerprints(
            configs,
            fingerprints_filename,
            zeta,
            dzetadr_forces,
            dzetadr_stress,
            fit_forces,
            fit_stress,
        )

        return fingerprints_filename

    def _dump_fingerprints(
        self,
        configs,
        fname,
        all_zeta,
        all_dzetadr_forces,
        all_dzetadr_stress,
        fit_forces,
        fit_stress,
    ):
        """
        Dump fingerprints to a pickle file.
        """

        logger.info(f"Pickling fingerprints to `{fname}`")

        create_directory(fname, is_directory=False)

        # remove it, because we use append mode for the file below
        fname = to_path(fname)
        if fname.exists():
            fname.unlink()

        with open(fname, "ab") as f:
            for i, conf in enumerate(configs):
                if i % 100 == 0:
                    logger.info(f"Processing configuration: {i}.")

                if all_zeta is None:
                    zeta, dzetadr_f, dzetadr_s = self.transform(
                        conf, fit_forces, fit_stress
                    )
                else:
                    zeta = all_zeta[i]
                    dzetadr_f = all_dzetadr_forces[i]
                    dzetadr_s = all_dzetadr_stress[i]

                # centering and normalization
                if self.normalize:
                    zeta = (zeta - self.mean) / self.stdev
                    if fit_forces or fit_stress:
                        stdev_3d = np.atleast_3d(self.stdev)
                    if fit_forces:
                        dzetadr_f = dzetadr_f / stdev_3d
                    if fit_stress:
                        dzetadr_s = dzetadr_s / stdev_3d

                # pickling data
                zeta = np.asarray(zeta, self.dtype)
                energy = np.asarray(conf.energy, self.dtype)
                if fit_forces:
                    dzetadr_f = np.asarray(dzetadr_f, self.dtype)
                    forces = np.asarray(conf.forces, self.dtype)
                if fit_stress:
                    dzetadr_s = np.asarray(dzetadr_s, self.dtype)
                    stress = np.asarray(conf.stress, self.dtype)
                    volume = np.asarray(conf.get_volume(), self.dtype)

                example = {"configuration": conf, "zeta": zeta, "energy": energy}
                if fit_forces:
                    example["dzetadr_forces"] = dzetadr_f
                    example["forces"] = forces
                if fit_stress:
                    example["dzetadr_stress"] = dzetadr_s
                    example["stress"] = stress
                    example["volume"] = volume

                pickle.dump(example, f)

        logger.info(f"Pickle {len(configs)} configurations finished.")

    def _calc_zeta_dzetadr(self, configs, fit_forces, fit_stress, nprocs=1):
        """
        Calculate the fingerprints and maybe its gradients w.r.t the atomic coords.
        """
        if nprocs == 1:
            zeta = []
            dzetadr_forces = []
            dzetadr_stress = []
            for conf in configs:
                z, dzdr_f, dzdr_s = self.transform(conf, fit_forces, fit_stress)
                zeta.append(z)
                dzetadr_forces.append(dzdr_f)
                dzetadr_stress.append(dzdr_s)
        else:
            rslt = parallel.parmap1(
                self.transform, configs, fit_forces, fit_stress, nprocs=nprocs
            )
            zeta = [pair[0] for pair in rslt]
            dzetadr_forces = [pair[1] for pair in rslt]
            dzetadr_stress = [pair[2] for pair in rslt]

        return zeta, dzetadr_forces, dzetadr_stress

    def _welford_mean_and_stdev(self, configs: List[Configuration]):
        """
        Compute the mean and standard deviation of fingerprints.

        This running mean and standard method proposed by Welford is memory-efficient.
        Besides, it outperforms the naive method from suffering numerical instability for
        large dataset.

        Args:
            configs: Dataset class:`~kliff.dataset.Configuration`.

        See Also:
            https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance
        """

        # number of features
        conf = configs[0]
        zeta, _, _ = self.transform(conf, fit_forces=False, fit_stress=False)
        size = zeta.shape[1]

        # Welford's method
        n = 0
        mean = np.zeros(size)
        M2 = np.zeros(size)
        for i, conf in enumerate(configs):
            zeta, _, _ = self.transform(conf, fit_forces=False, fit_stress=False)
            for row in zeta:
                n += 1
                delta = row - mean
                mean += delta / n
                delta2 = row - mean
                M2 += delta * delta2

            if i % 100 == 0:
                logger.info("Processing configuration: {i}.")
                sys.stdout.flush()

        stdev = np.sqrt(M2 / n)  # not the unbiased

        return mean, stdev

    def transform(
        self, conf: Configuration, fit_forces: bool = False, fit_stress: bool = False
    ) -> Tuple[np.ndarray, Optional[np.ndarray], Optional[np.ndarray]]:
        """
        Transform atomic coords to atomic environment descriptor values.

        Args:
            conf: atomic configuration
            fit_forces: Whether to fit forces, so as to compute gradients of fingerprints
                w.r.t. coords
            fit_stress: Whether to fit stress, so as to compute gradients of fingerprints
                w.r.t. coords

        Returns:
            zeta: Descriptor values. 2D array with shape (num_atoms, num_descriptors),
                where num_atoms is the number of atoms in the configuration, and
                num_descriptors is the size of the descriptor vector (depending on the
                choice of the hyperparameters).
            dzeta_dr: Gradient of the descriptor w.r.t. atomic coordinates. 4D array if
                grad is `True`, otherwise `None`. Shape: (num_atoms, num_descriptors,
                num_atoms, 3), where num_atoms and num_descriptors has the same meanings
                as described in zeta, and 3 denotes the 3D space for the Cartesian
                coordinates.
            dzeta_ds: Gradient of the descriptor w.r.t. virial stress component. 2D
                array of shape (num_atoms, num_descriptors, 6), where num_atoms and
                num_descriptors has the same meanings as described in zeta,
                and 6 denote the virial stress component in Voigt notation, see
                https://en.wikipedia.org/wiki/Voigt_notation
        """
        raise NotImplementedError

    def write_kim_params(
        self, path: Union[Path, str], fname: str = "descriptor.params"
    ):
        """
        Write descriptor info for KIM model.

        Args:
            path: Directory Path to write the file.
            fname: Name of the file.
        """
        raise NotImplementedError

    def get_size(self):
        """Return the size of the descriptor vector."""
        if self.mean is not None:
            size = len(self.mean)
        else:
            size = self.size
        return size

    def get_mean(self):
        """Return a list of the mean of the fingerprints."""
        return self.mean.copy()

    def get_stdev(self):
        """Return a list of the standard deviation of the fingerprints."""
        return self.stdev.copy()

    def get_dtype(self):
        """Return the data type of the fingerprints."""
        return self.dtype

    def get_cutoff(self):
        """Return the name and values of cutoff."""
        return self.cut_name, self.cut_dists

    def get_hyperparams(self):
        """Return the hyperparameters of descriptors."""
        return self.hyperparams

    def state_dict(self) -> Dict[str, Any]:
        """
        Return the state dict of the descriptor.
        """
        data = {"mean": self.mean, "stdev": self.stdev, "size": self.size}

        return data

    def load_state_dict(self, data: Dict[str, Any]):
        """
        Load state dict of a descriptor.

        Args:
            data: state dict to load.
        """
        try:
            mean = data["mean"]
            stdev = data["stdev"]
            size = data["size"]
        except Exception as e:
            raise DescriptorError(f"Corrupted state dict for descriptor: {str(e)}")

        # more checks on data integrity
        if mean is not None and stdev is not None and size is not None:
            if len(mean.shape) != 1 or mean.shape[0] != size:
                raise DescriptorError(f"Corrupted descriptor mean.")

            if len(stdev.shape) != 1 or stdev.shape[0] != size:
                raise DescriptorError("Corrupted descriptor standard deviation.")

        self.mean = mean
        self.stdev = stdev
        self.size = size


def load_fingerprints(path: Union[Path, str]):
    """
    Read preprocessed fingerprints from file.

    This is the reverse operation of Descriptor._dump_fingerprints.

    Args:
        path: Path to the pickled data file.

    Returns:
        Fingerprints
    """
    data = []
    with open(path, "rb") as f:
        try:
            while True:
                x = pickle.load(f)
                data.append(x)
        except EOFError:
            pass
        except Exception as e:
            raise DescriptorError(f"Cannot load fingerprints from `{path}`. {str(e)}")

    return data


def generate_full_cutoff(cutoff):
    """
    Generate a full binary cutoff dictionary.

    For species pair `S1-S2` in the ``cutoff`` dictionary, add key `S2-S1` to it, with
    the same value as `S1-S2`.

    Args:
        cutoff: Cutoff dictionary with key of the form ``A-B`` where ``A`` and ``B``
            are atomic species, and value should be a float.

    Returns:
        A dictionary with all combination of species as keys.

    Example:
    >>> cutoff = {'C-C': 4.0, 'C-H':3.5}
    >>> generate_full_cutoff(cutoff)
        {'C-C': 4.0, 'C-H':3.5, 'H-C':3.5}
    """
    rcut2 = dict()
    for key, val in cutoff.items():
        s1, s2 = key.split("-")
        if s1 != s2:
            reverse_key = s2 + "-" + s1
            if reverse_key in cutoff and cutoff[reverse_key] != val:
                raise Exception(
                    'Corrupted cutoff dictionary. cutoff["{0}-{1}"] != '
                    'cutoff["{1}-{0}"].'.format(s1, s2)
                )
            else:
                rcut2[reverse_key] = val
    # merge
    rcut2.update(cutoff)

    return rcut2


def generate_unique_cutoff_pairs(cutoff):
    """
    Generate a full binary cutoff dictionary.

    For species pair `S1-S2` in the ``cutoff`` dictionary, remove key `S2-S1` from it if
    `S1` is different from `S2`.

    Args:
        cutoff: Cutoff dictionary with key of the form ``A-B`` where ``A`` and ``B``
            are atomic species, and value should be a float.

    Returns:
        A dictionary with unique species pair as keys.

    Example:
    >>> cutoff = {'C-C': 4.0, 'C-H':3.5, 'H-C':3.5}
    >>> generate_unique_cutoff_pairs(cutoff)
        {'C-C': 4.0, 'C-H':3.5}
    """
    rcut2 = dict()
    for key, val in cutoff.items():
        s1, s2 = key.split("-")
        reverse_key = s2 + "-" + s1
        if key not in rcut2 and reverse_key not in rcut2:
            rcut2[key] = val

    return rcut2


def generate_species_code(cutoff):
    """
    Generate species code info from cutoff dictionary.

    Args:
        cutoff: Cutoff dictionary with key of the form ``A-B`` where ``A`` and ``B``
            are atomic species, and value should be a float.

    Returns:
        species_code: A dictionary of species and the integer code (starting from 0),
            with keys the species in ``cutoff`` keys, and values integer code for species.

    Example:
    >>> cutoff = {'C-C': 4.0, 'C-H':3.5}
    >>> generate_species_code(cutoff)
        {'C':0, 'H':1}
    """
    species = set()
    for key in cutoff:
        s1, s2 = key.split("-")
        species.update([s1, s2])
    species = list(species)
    species_code = {s: i for i, s in enumerate(species)}
    return species_code


class DescriptorError(Exception):
    def __init__(self, msg):
        super(DescriptorError, self).__init__(msg)
        self.msg = msg



from typing import Dict, List

import numpy as np

from kliff.models.parameter import Parameter


class ParameterTransform:
    """
    Abstract class to transform parameters and inverse transform it back.

    Subclass can implement
        - transform
        - inverse_transform
    """

    def transform(self, model_params: Dict[str, Parameter]) -> Dict[str, Parameter]:
        return model_params

    def inverse_transform(
        self, model_params: Dict[str, Parameter]
    ) -> Dict[str, Parameter]:
        return model_params

    def __call__(self, model_params: Dict[str, Parameter]) -> Dict[str, Parameter]:
        return self.transform(model_params)


class LogParameterTransform(ParameterTransform):
    """
    Transform parameters to a log space and transform it back.

    Args:
        param_names: names of the parameters to do the transformation; can be a
            subset of all the parameters.
    """

    def __init__(self, param_names: List[str]):
        super().__init__()
        self.param_names = param_names

    def transform(self, model_params: Dict[str, Parameter]) -> Dict[str, Parameter]:
        for name in self.param_names:
            p = model_params[name]
            p_val = np.asarray(p.value)

            assert np.asarray(p_val).min() >= 0, (
                f"Cannot log transform parameter `{name}`, got negative values: "
                f"{p_val}"
            )
            p.value = np.log(p_val)

        return model_params

    def inverse_transform(
        self, model_params: Dict[str, Parameter]
    ) -> Dict[str, Parameter]:
        for name in self.param_names:
            p = model_params[name]
            p_val = np.asarray(p.value)
            p.value = np.exp(p_val)

        return model_params

    def __call__(self, model_params: Dict[str, Parameter]) -> Dict[str, Parameter]:
        return self.transform(model_params)



from distutils.sysconfig import get_config_vars
from pathlib import Path

from setuptools import Extension, find_packages, setup

# remove `-Wstrict-prototypes' that is for C not C++
cfg_vars = get_config_vars()
for key, value in cfg_vars.items():
    if type(value) == str and "-Wstrict-prototypes" in value:
        cfg_vars[key] = value.replace("-Wstrict-prototypes", "")


class get_pybind11_includes:
    """
    Helper class to determine the pybind11 include path

    The purpose of this class is to postpone importing pybind11 until it is actually
    installed, so that the ``get_include()`` method can be invoked.

    see:
    https://github.com/pybind/python_example/blob/master/setup.py
    https://github.com/pybind/python_example/issues/32
    """

    def __str__(self):
        import pybind11

        return pybind11.get_include()


def get_includes():
    return [get_pybind11_includes()]


def get_extra_compile_args():
    return ["-std=c++11"]


sym_fn = Extension(
    "kliff.descriptors.symmetry_function.sf",
    sources=[
        "kliff/descriptors/symmetry_function/sym_fn_bind.cpp",
        "kliff/descriptors/symmetry_function/sym_fn.cpp",
        "kliff/descriptors/symmetry_function/helper.cpp",
    ],
    include_dirs=get_includes(),
    extra_compile_args=get_extra_compile_args(),
    language="c++",
)

bispectrum = Extension(
    "kliff.descriptors.bispectrum.bs",
    sources=[
        "kliff/descriptors/bispectrum/bispectrum_bind.cpp",
        "kliff/descriptors/bispectrum/bispectrum.cpp",
        "kliff/descriptors/bispectrum/helper.cpp",
    ],
    include_dirs=get_includes(),
    extra_compile_args=get_extra_compile_args(),
    language="c++",
)

neighlist = Extension(
    "kliff.neighbor.neighlist",
    sources=[
        "kliff/neighbor/neighbor_list.cpp",
        "kliff/neighbor/neighbor_list_bind.cpp",
    ],
    include_dirs=get_includes(),
    extra_compile_args=get_extra_compile_args(),
    language="c++",
)


def get_version():
    fname = Path(__file__).parent.joinpath("kliff", "__init__.py")
    with open(fname) as f:
        for line in f:
            line = line.strip()
            if "__version__" in line:
                v = line.split("=")[1]
                # stripe white space, and ' or " in string
                if "'" in v:
                    version = v.strip("' ")
                elif '"' in v:
                    version = v.strip('" ')
                break
    return version


def get_readme():
    fname = Path(__file__).parent.joinpath("README.md")
    with open(fname, "r") as f:
        readme = f.read()
    return readme


setup(
    name="kliff",
    version=get_version(),
    packages=find_packages(),
    ext_modules=[sym_fn, bispectrum, neighlist],
    install_requires=[
        "requests",
        "pybind11",
        "scipy",
        "pyyaml",
        "monty",
        "loguru",
    ],
    extras_require={
        "test": ["pytest"],
    },
    entry_points={"console_scripts": ["kliff = kliff.cmdline.cli:main"]},
    author="Mingjian Wen",
    author_email="wenxx151@gmail.com",
    url="https://github.com/openkim/kliff",
    description="KLIFF: KIM-based Learning-Integrated Fitting Framework",
    long_description=get_readme(),
    long_description_content_type="text/markdown",
    classifiers=[
        "License :: OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)",
        "Operating System :: OS Independent",
    ],
    zip_safe=False,
)



import copy
from typing import Any, Dict, Optional

import numpy as np
from loguru import logger


class Weight:
    """Base class for weight.

    This class mimics the behavior  provided by the ``residual_data`` argument from the
    previous version.
    Subclass can implement
        - compute_weight

    Parameters
    ----------
    config_weight: float (optional)
        The weight for the entire configuration
    energy_weight: float (optional)
        The weight for the energy
    forces_weight: float (optional)
        The weight for the forces
    stress_weight: float (optional)
        The weight for the stress
    """

    def __init__(
        self,
        config_weight: float = 1.0,
        energy_weight: float = 1.0,
        forces_weight: float = 1.0,
        stress_weight: float = 1.0,
    ):
        self._config_weight = config_weight
        self._energy_weight = energy_weight
        self._forces_weight = forces_weight
        self._stress_weight = stress_weight

    def compute_weight(self, config):
        self._check_compute_flag(config)

    @property
    def config_weight(self):
        return self._config_weight

    @property
    def energy_weight(self):
        return self._energy_weight

    @property
    def forces_weight(self):
        return self._forces_weight

    @property
    def stress_weight(self):
        return self._stress_weight

    def _check_compute_flag(self, config):
        """
        Check whether compute flag correctly set when the corresponding weight in
        residual data is 0.
        """
        ew = self.energy_weight
        fw = self.forces_weight
        sw = self.stress_weight
        msg = (
            '"{0}_weight" are near zero. Seems you do not want to use {0} in the '
            'fitting. You can set "use_{0}" in "calculator.create()" to "False" to speed'
            "up the fitting."
        )

        # If the weight are really small, but not zero, then warn the user. Zero weight
        # usually means that the property is used.
        if config._energy is not None and np.all(ew < 1e-12):
            logger.warning(msg.format("energy", ew))
        if config._forces is not None and np.all(fw < 1e-12):
            logger.warning(msg.format("forces", fw))
        if config._stress is not None and np.all(sw < 1e-12):
            logger.warning(msg.format("stress", sw))


class MagnitudeInverseWeight(Weight):
    r"""Non-uniform weight that is computed from the data. The calculation follows
    Lenosky et al. (1997), with some modification in notation,

    ..math:
       \frac{1}{w_m^2} = c_1^2 + c_2^2 \Vert f_m \Vert ^2.

    Parameters
    ----------
    config_weight: float (optional)
        The weight for the entire configuration
    weight_params: dict (optional)
        A dictionary containing parameters c1 and c2 for calculating the
        weight of each property. The supported key value pairs are:
        - energy_weight_params: float or array-like (default: [1.0, 0.0])
        - forces_weight_params: float or array-like (default: [1.0, 0.0])
        - stress_weight_params: float or array-like (default: [1.0, 0.0])
        If a float is given, this number will be used to set c1, while c2 is set to
        zero. If an array-like with 2 elements is given, it should contain c1 as the
        first element and c2 as the second element.

    References
    ----------
    .. [Lenosky1997] T. J. Lenosky et al., “Highly optimized tight-binding model of
       silicon,” Phys. Rev. B, vol. 55, no. 3, pp. 15281544, Jan. 1997, doi:
       10.1103/PhysRevB.55.1528.

    """

    # Default parameters
    default_weight_params = {
        "energy_weight_params": [1.0, 0.0],
        "forces_weight_params": [1.0, 0.0],
        "stress_weight_params": [1.0, 0.0],
    }

    def __init__(
        self,
        config_weight: float = 1.0,
        weight_params: Optional[Dict[str, Any]] = None,
    ):
        self._config_weight = config_weight
        # Initiate the weight values. They will be changed latter.
        self._energy_weight = 0.0
        self._forces_weight = 0.0
        self._stress_weight = 0.0

        self._weight_params = self._check_weight_params(
            weight_params, self.default_weight_params
        )

    def compute_weight(self, config):
        """
        Compute the weights of the energy, forces, and stress data.
        """
        energy = config._energy
        forces = config._forces
        stress = config._stress

        # Energy
        if energy is not None:
            # Use the absolute value of the energy
            energy_norm = np.abs(energy)
            self._energy_weight = self._compute_weight_one_property(
                energy_norm, self._weight_params["energy_weight_params"], "energy"
            )
        # Forces
        if forces is not None:
            # Use the magnitude of the force vector
            forces_norm = np.linalg.norm(forces, axis=1)
            self._forces_weight = np.repeat(
                self._compute_weight_one_property(
                    forces_norm, self._weight_params["forces_weight_params"], "forces"
                ),
                3,
            )
        # Stress
        if stress is not None:
            # Use the Frobenius norm of the stress tensor
            normal_stress_norm = np.linalg.norm(stress[:3])
            shear_stress_norm = np.linalg.norm(stress[3:])
            stress_norm = np.sqrt(normal_stress_norm**2 + 2 * shear_stress_norm**2)
            self._stress_weight = self._compute_weight_one_property(
                stress_norm, self._weight_params["stress_weight_params"], "stress"
            )

        self._check_compute_flag(config)

    @staticmethod
    def _compute_weight_one_property(data_norm, property_weight_params, property_type):
        """
        Compute the weight based for one property.
        """
        c1, c2 = property_weight_params
        sigma = np.linalg.norm([c1, (c2 * data_norm)])
        weight = 1 / sigma
        if np.any(sigma < 1e-12):
            logger.warning(
                f"Found near zero inverse {property_type} weight. Be aware that some "
                f"{property_type} data might be overweight."
            )
        return weight

    @staticmethod
    def _check_weight_params(weight_params: Dict[str, Any], default: Dict[str, Any]):
        """
        Check the weight parameters and set it to the needed format, i.e., list with 2
        elements for each property.
        """
        if weight_params is not None:
            for key, value in weight_params.items():
                if key not in default:
                    raise WeightError(
                        f"Expect the keys of `weight_params` to be one or combinations "
                        f"of {', '.join(default.keys())}; got {key}. "
                    )
                else:
                    if np.ndim(value) == 0:
                        # If there is only a number given, use it to set c1
                        default[key][0] = value
                    elif np.ndim(value) == 1 and len(value) == 2:
                        # To set c1 and c2, a list with 2 elements need to be passed in
                        default[key] = value
                    else:
                        raise WeightError(
                            "Expect a float or a list of floats with format [c1, c2]"
                        )
        return default


class WeightError(Exception):
    def __init__(self, msg):
        super(WeightError, self).__init__(msg)
        self.msg = msg
